//
// antsd:  Main Job Daemon Code
//
// $Id: antsd.cc,v 1.7 2001/03/11 20:41:15 jakob Exp $
//
//  Copyright (C) 2000 Jakob Oestergaard <jakob@unthought.net>
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later
//  version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include "antsd.hh"
#include "sock.hh"
#include "jobnet.hh"
#include "cluster.hh"
#include "processes.hh"

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <errno.h>
#include <signal.h>
#include <sys/stat.h>

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_SYSLOG_H
# include <syslog.h>
#endif

#include <errno.h>
#include <signal.h>
#include <sys/stat.h>

#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif

#include <list>
#include <string>
#include <algorithm>
#include <iostream>

struct recall_t {
  int times_left;
  unsigned delay;
  struct timeval at;
  bool(*func)(int,unsigned);
  unsigned usr_data;
};

std::list<recall_t> recall_list;
int debug_mode = 0;

void usage();
void treat_recall_routines();
void sigterm_handler(int);

bool shutdown_issued = false;

int main(int argc, char** argv)
{
  bool do_exit = false;

  struct sigaction sigpipe_act;
  memset(&sigpipe_act, 0, sizeof(sigpipe_act));
  sigpipe_act.sa_handler = SIG_IGN;
  sigaction(SIGPIPE, &sigpipe_act, 0);

  struct sigaction sigterm_act;
  memset(&sigterm_act, 0, sizeof(sigterm_act));
  sigterm_act.sa_handler = sigterm_handler;
  sigaction(SIGTERM, &sigterm_act, 0);

  char** argp = argv + 1;
  while(*argp) {
    if(!strcmp(*argp, "-h")) usage();
    else if(!strcmp(*argp, "-d")) debug_mode++;
    else usage();
    argp++;
  }

  if(!debug_mode) {
    openlog("antsd", LOG_NDELAY, LOG_DAEMON);

    int rc;
    if((rc = fork())) {
      if(rc < 0)
	exit_fatal("Cannot fork");
      //
      // We're the parent
      //
      exit(0);
    }
    //
    // We're the child - just shut ourselve up.
    // After all, children should be seen not heard  ;)
    //
    int nullfd = open("/dev/null", O_RDWR);
    if(nullfd < 0) exit_fatal("Couldn't open /dev/null");
    dup2(nullfd, 0);
    dup2(nullfd, 1);
    dup2(nullfd, 2);
    setsid();
  }

  try {
    initialize_jobnet();
    initialize_cluster();
    initialize_processes();
    
    log_info("Started.");

    while(!do_exit) {
      int max_fd = -1;
      fd_set read, except, write;
      FD_ZERO(&read);
      FD_ZERO(&except);
      FD_ZERO(&write);

      if(shutdown_issued) {
	if(jobsock != -1) shutdown_jobnet();
      }
    
      if(jobsock == -1 && !shutdown_issued)
        setup_jobnet();

      if(jobsock != -1) {
        FD_SET(jobsock, &read);
        FD_SET(jobsock, &except);
        max_fd = std::max(max_fd, jobsock);
      }

      if(localsock == -1 && !shutdown_issued)
        setup_sock();

      if(localsock != -1) {
        FD_SET(localsock, &read);
        FD_SET(localsock, &except);
        max_fd = std::max(max_fd, localsock);
      }

      if(process_sock != -1) {
        FD_SET(process_sock, &read);
        FD_SET(process_sock, &except);
        max_fd = std::max(max_fd, process_sock);
      }

      if(max_fd == -1) {
        log_info("Both sock and jobnet is down, will retry.");
        sleep(RETRY_DELAY);
        continue;
      }

      //
      // Maintain timers
      //
      struct timeval tvnow;
      gettimeofday(&tvnow, 0);    
      const time_t now = tvnow.tv_sec;
      struct timeval timeout;
      timeout.tv_usec = 0;

      static time_t last_jobnet_update = 0;
      if(last_jobnet_update + JOBNET_UPDATE_INTERVAL <= now) {
        maintain_jobnet();
        last_jobnet_update = now;
      }
      timeout.tv_sec = last_jobnet_update + JOBNET_UPDATE_INTERVAL - now;
      static time_t last_cluster_update = 0;
      if(last_cluster_update + CLUSTER_UPDATE_INTERVAL <= now) {
        maintain_cluster();
        last_cluster_update = now;
      }
      timeout.tv_sec = std::min(timeout.tv_sec, last_cluster_update + CLUSTER_UPDATE_INTERVAL - now);

      //
      // Iterate over all spawned sub-processes awaiting authentication completion
      //
      for(std::map<int,process>::const_iterator pi = process_table.begin();
	  pi != process_table.end(); pi++) {
	if(pi->second.auth_ok) continue;
	time_t sec_diff = pi->second.spawn_time + JOBCONN_AUTH_TIMEOUT - now;
	// Otherwise insert in timeout
	if(sec_diff * 1000 < (timeout.tv_sec * 1000 + timeout.tv_usec / 1000)) {
	  timeout.tv_sec = sec_diff;
	  timeout.tv_usec = 0;
	}
      }

      //
      // Iterate over recall timers
      //
      for(std::list<recall_t>::iterator ri = recall_list.begin();
	  ri != recall_list.end(); ri++) {
	int msec_diff = (ri->at.tv_sec - tvnow.tv_sec) * 1000
	  + (ri->at.tv_usec - tvnow.tv_usec) / 1000;
	int old_diff = timeout.tv_sec * 1000 + timeout.tv_usec / 1000;

	if(msec_diff < old_diff) {
	  timeout.tv_sec = msec_diff / 1000;
	  timeout.tv_usec = msec_diff * 1000;
	}
      }

      //
      // Fixup timeout
      //
      while(timeout.tv_usec < 0) {
	timeout.tv_sec --;
	timeout.tv_usec += 1000000L;
      }
      if(timeout.tv_sec < 0) {
	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
      }

      //
      // Unconditionally maintain processes
      //
      maintain_processes();

      //
      // Insert sub-process handler pipes
      //
      insert_process_pipe_set(read, max_fd);
      insert_process_pipe_set(except, max_fd);

      // Insert local AF_UNIX connections
      insert_local_readers_set(read, max_fd);
      insert_local_readers_set(except, max_fd);
      insert_local_writers_set(write, max_fd);

      //
      // select()
      //
      int rc;
      while(-1 == (rc = select(max_fd+1, &read, &write, &except, &timeout)))
	if(errno == EINTR) maintain_processes();
	else break;
      
      if(rc == -1) {
        log_info("select() failed. Will retry. (" 
		 + std::string(strerror(errno)) + ")");
        sleep(RETRY_DELAY);
        continue;
      }

      treat_recall_routines();

      if(!rc)
        continue;

      if(localsock != -1 && FD_ISSET(localsock, &read))
        read_sock();
      if(jobsock != -1 && FD_ISSET(jobsock, &read))
        read_jobnet();

      if(localsock != -1 && FD_ISSET(localsock, &except))
        except_sock();
      if(jobsock != -1 && FD_ISSET(jobsock, &except))
        except_jobnet();

      // Read from local connections
      read_local_set(read);
      read_local_set(except);

      // Write pending stuff to local connections
      write_local_set(write);

      // Read from processes
      read_processes_pipe_set(read);
      read_processes_pipe_set(except);

    }
  } catch(internalerror e) {
    log_info("Internal error: \"" + e.message + "\"");
  } catch(...) {
    log_info("We caught an unexpected exception. Exiting.");
  }
  if(!debug_mode)
    closelog();

  return 0;
}


void log_info(const std::string& s, unsigned lev)
{
  if(debug_mode > 0)
    std::cerr << s << std::endl;
  else if(lev)
    syslog(LOG_INFO, s.c_str(), s.length());
}

void exit_fatal(const std::string& s)
{
  log_info(s);
  if(debug_mode) abort();
  exit(1);
}

void usage()
{
  std::cerr << "Usage:" << std::endl
	    << "help:       antsd -h" << std::endl
	    << "debug:      antsd -d" << std::endl
	    << "louder:     antsd -d -d" << std::endl
	    << "normal:     antsd" << std::endl;
  exit(0);
}

void add_recall_routine(bool(*f)(int,unsigned), unsigned interv, unsigned retries, 
			unsigned usr_data)
{
  recall_t rnew;
  rnew.times_left = retries;
  rnew.delay = interv;
  gettimeofday(&rnew.at, 0);
  rnew.func = f;
  rnew.usr_data = usr_data;

  recall_list.push_back(rnew);
}

void treat_recall_routines()
{
  struct timeval tvnow;
  gettimeofday(&tvnow, 0);    

  for(std::list<recall_t>::iterator ri = recall_list.begin();
	ri != recall_list.end(); ) {
    std::list<recall_t>::iterator next = ri; next++;
    int msec_diff = (ri->at.tv_sec - tvnow.tv_sec) * 1000
      + (ri->at.tv_usec - tvnow.tv_usec) / 1000;
    if(msec_diff <= 0) {
      //
      // Call routine
      //
      ri->times_left--;
      bool remove = !ri->func(ri->times_left, ri->usr_data);

      if(remove || ri->times_left <= 0) {
	//
	// Remove from list
	//
	recall_list.erase(ri);
      } else {
	//
	// Recalc timer
	//
	ri->at.tv_usec = tvnow.tv_usec + ri->delay*1000;
	ri->at.tv_sec = tvnow.tv_sec + ri->delay/1000;
      }

    }
    ri = next;
  }
}

void remove_recall_routine(bool(*f)(int,unsigned), unsigned usr_data)
{
  for(std::list<recall_t>::iterator ri = recall_list.begin();
      ri != recall_list.end(); ri++)
    if(ri->func == f && ri->usr_data == usr_data) {
      recall_list.erase(ri);
      log_info("Killed recall");
      break;
    }
}

void sigterm_handler(int s)
{
  //
  // When receiving a SIGTERM, we refuse new connections
  // but let our running processes finish...
  //
  shutdown_issued = true;
}
