//
// Implementation of cluster helper routines
//
// $Id: cluster.cc,v 1.19 2001/03/11 20:41:15 jakob Exp $
//
//  Copyright (C) 2000 Jakob Oestergaard <jakob@unthought.net>
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later
//  version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "cluster.hh"
#include "processes.hh"
#include "jobnet.hh"
#include "antsd.hh"
#include "sock.hh"

#include <list>
#include <string>
#include <set>
#include <map>
#include <algorithm>
#include <fstream>

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <netdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif

//
// We must know who we are
//
address_t local_address;

//
// The set of hosts in the cluster, mapping into their metrics
//

struct load_data {
  jobt_limits limits;
  jobt_metrics metrics;
  double goodness() const {
    if(limits.max_instances <= metrics.instances + metrics.preallocations)
      return 0;
    //
    // the +1 below will guard for div. by zero and account for new instance
    //
    return double(limits.preferred_instances)
      / double(metrics.preallocations + metrics.instances + 1);
  }
};

std::map<address_t, std::map<std::string, load_data> > hostmetrics;

std::map<std::string, std::multiset<time_t> > prealloc_slots;

//
// Seq# -> allocation_time mapping
//
struct seq_alloc {
  time_t atime;
  address_t host;
  std::string type;
  std::string username;
};
std::map<unsigned,seq_alloc> used_seqs;

//
// Recently received allocation requests list
//
struct serial_cache_t {
  time_t atime;
  char * buf;
  int len;
};
// Request serial number -> hostname ->  reply
std::map<unsigned,std::map<address_t, serial_cache_t> > serial_cache;


void update_local_metrics(bool force)
{
  std::map<address_t, std::map<std::string, load_data> >::iterator hi = hostmetrics.find(local_address);
  if(hi == hostmetrics.end())
    throw internalerror("Cannot find myself in get_local_metrics... This can't happen");

  //
  // We conditionally update the more expensive stuff
  //  
  static time_t last_local_update = 0;
  time_t now = time(0);
  if(force || last_local_update + LOCAL_METRIC_UPDATE_INTERVAL < now) {
    //
    // Update local metrics
    //
    if(force)
      log_info("Obeying forced local metric update");
    else
      log_info("Will update local metrics");

    for(std::map<std::string, load_data>::iterator ti = hi->second.begin();
        ti != hi->second.end(); ti++)
      ti->second.metrics.zero();

    std::map<std::string, jobt_metrics> running_metrics;

    //
    // Update metrics wrt. currently running jobs
    //
    for(std::map<int,process>::const_iterator pi = process_table.begin();
        pi != process_table.end(); pi++)
      running_metrics[pi->second.type] += pi->second.get_jobt_metrics();

    for(std::map<std::string, jobt_metrics>::const_iterator ri = running_metrics.begin();
        ri != running_metrics.end(); ri++)
      hi->second[ri->first].metrics = ri->second;


    last_local_update = now;
 
  }

  //
  // We always update the preallocations, becuase this is assumed by other parts of
  // the daemon and because it's really cheap.
  //
  for(std::map<std::string, std::multiset<time_t> >::const_iterator psi = prealloc_slots.begin();
      psi != prealloc_slots.end(); psi++) {
    hi->second[psi->first].metrics.preallocations = psi->second.size();
  }

  //
  // Call the UNIX socket handler for metric updates -
  // we might have some nifty applet showing what's happening...
  //
  for(std::map<std::string, load_data>::const_iterator lld = hi->second.begin();
      lld != hi->second.end(); lld++)
    notify_client_metrics_change(local_address, lld->first, 
				 lld->second.limits, lld->second.metrics);

}

void initialize_cluster()
{
  //
  // Find my own name... This is ugly code...
  //
  char myname[1024];
  int grc = gethostname(myname, 1024);
  if(grc)
    exit_fatal("Unable to look up my own name");
  struct hostent * hent = gethostbyname(myname);
  if(!hent)
    exit_fatal("Cannot look up my own host address");
  if(hent->h_length != sizeof(unsigned))
    exit_fatal("Local host address length mismatch!");
  unsigned * adr = reinterpret_cast<unsigned*>(hent->h_addr_list[0]);
  local_address.inet_port = JOBNET_PORT;
  local_address.inet_addr = *adr;

  //
  // Read in limits
  //
  ifstream configf(CONFIGFILE.c_str());
  while(configf) {
    //
    // Config file format:
    // <jobtype> <max_instances> <pref_instances>
    // 
    std::string type;
    load_data ldat;
    configf >> type 
            >> ldat.limits.max_instances
            >> ldat.limits.preferred_instances;
    if(!type.size()) continue;

    hostmetrics[local_address][type] = ldat;

    log_info("Read config for job type \"" + type + "\"");
  }

  if(!hostmetrics.size())
    exit_fatal("No metrics - nothing to do");
}

void maintain_cluster()
{
  const time_t now = time(0);

  //
  // Expire slots allocated for remote nodes
  //
  for(std::map<std::string,std::multiset<time_t> >::iterator ti = prealloc_slots.begin();
      ti != prealloc_slots.end(); ti++)
    for(std::multiset<time_t>::iterator si = ti->second.begin();
        si != ti->second.end(); ) 
      if(*si + CLUSTER_PREALLOC_SLOT_LIFE < now) {
        std::multiset<time_t>::iterator next = si; next++;
        ti->second.erase(si);
        log_info("Expired allocated job-slot");
        si = next;
      } else break;

  //
  // Update local metrics, because they may have changed during the expire above
  //
  update_local_metrics(false);  // explicitly do _not_ force, as we call this often

  //
  // Expire old allocation requests cached
  //
  static time_t last_serial_cache_prune = 0;
  if(last_serial_cache_prune + CLUSTER_SERIAL_CACHE_LIFE < now) {
    std::map<unsigned,std::map<address_t,serial_cache_t> >::iterator i = serial_cache.begin();
    while(i != serial_cache.end()) {
      std::map<unsigned,std::map<address_t,serial_cache_t> >::iterator next = i; next++;

      std::map<address_t,serial_cache_t>::iterator ii = i->second.begin();
      while(ii != i->second.end()) {
	std::map<address_t,serial_cache_t>::iterator inext = ii; inext++;
	if(ii->second.atime + CLUSTER_SERIAL_CACHE_LIFE < now) {
	  delete[] ii->second.buf;
	  i->second.erase(ii);
	}
	ii = inext;
      }

      if(i->second.size())
	serial_cache.erase(i);
      i = next;
    }
    last_serial_cache_prune = now;
  }
}


bool suggest_job_allocation(const std::string& type, address_t& best_host)
{
  //
  // Iterate thru all hosts, considering our idea of their current
  // job loads.
  //
  double best_goodness = -1;

  for(std::map<address_t, std::map<std::string, load_data> >::const_iterator hi = hostmetrics.begin();
      hi != hostmetrics.end(); hi++) {
    std::map<std::string, load_data>::const_iterator jobti = hi->second.find(type);
    if(jobti == hi->second.end()) continue;

    double goodness = jobti->second.goodness();

    if(goodness > best_goodness) {
      best_host = hi->first;
      best_goodness = goodness;
    }
  }

  if(best_goodness <= 0) {
    log_info("Sorry, no best host found for type \"" + type + "\"");
    //
    // Request metrics from other hosts to enhance the chance
    // of a success next time
    //
    static time_t throttle_rnfa = 0;
    time_t now = time(0);
    if(throttle_rnfa + CLUSTER_METRIC_REQUEST_INTERVAL < now) {
      request_metrics_from_all();
      throttle_rnfa = now;
    }

    return false;
  }

  return true;
}


bool can_allocate(const std::string& type)
{
  //
  // Find limits for job type and get the current local
  // metrics for processes of that type
  //
  std::map<address_t, std::map<std::string, load_data> >::const_iterator hosti = hostmetrics.find(local_address);
  if(hosti == hostmetrics.end())
    throw internalerror("In can_allocate(): I can't even find myself... Going down now.");

  std::map<std::string, load_data>::const_iterator joblim = hosti->second.find(type);
  if(joblim == hosti->second.end()) {
    log_info("Got can_allocate() request for unknown job type \"" + type + "\"");
    return false;
  }

  return ! joblim->second.metrics.fills(joblim->second.limits);
}

bool is_serial_in_use(unsigned seq, const address_t& host)
{
  std::map<unsigned,std::map<address_t,serial_cache_t> >::const_iterator
    si = serial_cache.find(seq);
  if(si == serial_cache.end()) return false;

  std::map<address_t, serial_cache_t>::const_iterator ii = si->second.find(host);
  if(ii == si->second.end()) return false;

  return true;
}

void record_serial_in_use(unsigned serial, const address_t& host, 
			  const char * pkt, unsigned len)
{
  serial_cache_t newc;
  newc.len = len;
  newc.buf = new char[len];
  memcpy(newc.buf, pkt, len);
  newc.atime = time(0);
  serial_cache[serial][host] = newc;
}

unsigned fillin_serial_host_reply(unsigned serial, const address_t& host, char* buf)
{
  std::map<unsigned,std::map<address_t,serial_cache_t> >::const_iterator
    si = serial_cache.find(serial);
  if(si == serial_cache.end()) 
    throw internalerror("Cannot find host serial in serial cache");

  std::map<address_t, serial_cache_t>::const_iterator ii = si->second.find(host);
  if(ii == si->second.end())
    throw internalerror("Cannot find host address in serial cache");

  memcpy(buf, ii->second.buf, ii->second.len);
  return ii->second.len;
}


//
// We return 0 on error.
//
unsigned register_allocation(const std::string& type, const std::string& username, const address_t& host)
{
  std::map<address_t, std::map<std::string, load_data> >::const_iterator hosti = hostmetrics.find(local_address);
  if(hosti == hostmetrics.end())
    throw internalerror("In register_allocation(): I can't even find myself... Going down now.");

  std::map<std::string, load_data>::const_iterator joblim = hosti->second.find(type);
  if(joblim == hosti->second.end()) {
    log_info("Got register_allocation() request for unknown job type \"" + type
	     + "\" - peer \"" +  host.toString() + "\"is b0rken!");
    return 0;
  }

  //
  // The serial# must be unique and random.
  // A predictable serial number leads to easy remote-execution attacks
  //
  int rfile = open(RANDOMFILE.c_str(), O_RDONLY); // We use POSIX I/O here because SGI's libstdc++ is lacking features
  if(rfile < 0) 
    exit_fatal("Cannot open " + RANDOMFILE);
  unsigned serial;
  unsigned rd;
  do {
    rd = 0;
    do {
      int rc = read(rfile, reinterpret_cast<char*>(&serial) + rd, sizeof(serial)-rd);
      if(rc <= 0)
        exit_fatal("Cannot read from " + RANDOMFILE);
      rd += rc;
      if(rd > sizeof(serial))
        throw internalerror("Read read too much from " + RANDOMFILE);
    } while(rd != sizeof(serial));
  } while(!serial || used_seqs.find(serial) != used_seqs.end());
  close(rfile);

  seq_alloc sal;
  sal.host = host;
  sal.type = type;
  sal.atime = time(0);
  sal.username = username;
  
  //
  // Punch in
  //
  used_seqs.insert(make_pair(serial, sal));

  //
  // We update our number of preallocations and send back metrics to peer
  //
  prealloc_slots[type].insert(time(0));

  // needed for preallocations to take effect - do not force, only need preallocs.
  update_local_metrics(false);

  // caller of register allocation must call   send_metrics(host)   if host != local_address

  log_info("Registered allocation of job with type \"" + type + "\" to \"" +  host.toString() + "\"");

  return serial;
}


void update_metrics(const address_t& host, const std::string& type, const jobt_metrics& metrics)
{
  std::map<address_t, std::map<std::string, load_data> >::iterator hi = hostmetrics.find(host);
  if(hi == hostmetrics.end()) {
    log_info("Couldn't find host \"" +  host.toString() + "\" for metrics update");
    return;
  }
  
  std::map<std::string, load_data>::iterator ti = hi->second.find(type);
  if(ti == hi->second.end()) {
    log_info("Couldn't find type \"" + type + "\" in host \"" +  host.toString() + "\" metric table");
    //
    // Send a request for new limits to that host, since it obviously has
    // a job type we don't know about
    //
    request_new_limits(host);
    return;
  }

  ti->second.metrics = metrics;

  //
  // Call the UNIX socket handler for metric updates -
  // we might have some nifty applet showing what's happening...
  //
  notify_client_metrics_change(host, type, ti->second.limits, ti->second.metrics);

}

void invalidate_host_metrics(const address_t& host)
{
  std::map<address_t, std::map<std::string, load_data> >::iterator hi = hostmetrics.find(host);
  if(hi == hostmetrics.end()) {
    log_info("Couldn't find host \"" +  host.toString() + "\" for metrics invalidation", 1); // no big deal
    return;
  }
  hostmetrics.erase(hi);
}

void update_limits(const address_t& host, const std::string& type, const jobt_limits& limits)
{
  std::map<address_t, std::map<std::string, load_data> >::iterator hi = hostmetrics.find(host);
  if(hi == hostmetrics.end()) {
    log_info("Couldn't find host \"" +  host.toString() + "\" for limits update");
    return;
  }
  
  std::map<std::string, load_data>::iterator ti = hi->second.find(type);
  if(ti == hi->second.end()) {
    log_info("Received limits for previously unknown job type \"" + type
	     + "\" from \"" +  host.toString() + "\"");
    load_data new_entry;
    new_entry.metrics.invalidate();
    new_entry.limits = limits;
    hi->second.insert(make_pair(type, new_entry));
    return;
  }

  ti->second.limits = limits;
}

void get_local_metrics(std::map<std::string,jobt_metrics>& metrics)
{
  std::map<address_t, std::map<std::string, load_data> >::iterator hi = hostmetrics.find(local_address);
  if(hi == hostmetrics.end())
    throw internalerror("Cannot find myself in get_local_metrics... This can't happen");

  //
  // If it's too long ago we updated local metrics, do so now
  //
  update_local_metrics(false);

  metrics.clear();
  for(std::map<std::string, load_data>::const_iterator mi = hi->second.begin();
      mi != hi->second.end(); mi++)
    metrics.insert(make_pair(mi->first, mi->second.metrics));
}

void get_local_limits(std::map<std::string,jobt_limits>& limits)
{
  std::map<address_t, std::map<std::string, load_data> >::iterator hi = hostmetrics.find(local_address);
  if(hi == hostmetrics.end())
    throw internalerror("Cannot find myself in get_local_limits... This can't happen");

  limits.clear();
  for(std::map<std::string, load_data>::const_iterator mi = hi->second.begin();
      mi != hi->second.end(); mi++)
    limits.insert(make_pair(mi->first, mi->second.limits));
}

void initialize_load_data_for_host(const address_t& host)
{
  std::map<address_t, std::map<std::string, load_data> >::iterator hi = hostmetrics.find(host);
  if(hi == hostmetrics.end()) {
    log_info("Set up new load data record for remote host \"" +  host.toString() + "\"");
    hostmetrics.insert(make_pair(host, std::map<std::string,load_data>()));
    return;
  }
  log_info("Load data record for host \"" +  host.toString() + "\" already exists - will reset.");
  hi->second.clear();
}

void debug_print_load_data()
{
  for(std::map<address_t, std::map<std::string, load_data> >::iterator hi = hostmetrics.begin();
      hi != hostmetrics.end(); hi++) {
    for(std::map<std::string, load_data>::const_iterator mi = hi->second.begin();
        mi != hi->second.end(); mi++) {
      if(hi->first == local_address) std::cerr << "localhost:";
      else std::cerr << "remote(" << hi->first.toString() << "):";
      std::cerr << " jt:" << mi->first
                << " mi:" << mi->second.limits.max_instances
                << " pi:" << mi->second.limits.preferred_instances
                << " i:" << mi->second.metrics.instances
                << " p:" << mi->second.metrics.preallocations
		<< " g:" << mi->second.goodness()
                << std::endl;
    }
  }

}

std::string get_jobslot_username(unsigned seq)
{
  std::map<unsigned,seq_alloc>::const_iterator si =  used_seqs.find(seq);
  if(si == used_seqs.end()) throw internalerror("get_jobslot_username called with unknown seq#");
  return si->second.username;
}

void drop_prealloc_slot(const std::string& type)
{
  std::map<std::string, std::multiset<time_t> >::iterator psi =  prealloc_slots.find(type);
  if(psi == prealloc_slots.end()) return; // No big deal

  if(psi->second.size() == 0) return; // No big deal either

  psi->second.erase(psi->second.begin());  // Expire smallest time_t eg. oldest allocation

}

