//
// The Cluster definitions
//
// $Id: cluster.hh,v 1.13 2000/12/26 22:26:27 jakob Exp $
//
//  Copyright (C) 2000 Jakob Oestergaard <jakob@unthought.net>
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later
//  version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#ifndef __CLUSTER_HH
#define __CLUSTER_HH

#include "jobnet.hh"

#include <iostream>
#include <string>
#include <map>

//
// We identify a job-type with a string.
// Each node has specific limits for each job type.
//

struct jobt_limits {
  unsigned  max_instances;
  unsigned  preferred_instances;
  jobt_limits() : max_instances(0), 
    preferred_instances(0) {}
};


//
// Each node has a metric for each job type
//

struct jobt_metrics {
  unsigned instances;
  unsigned preallocations;
  jobt_metrics() : instances(0), preallocations(0) {}
  void zero() {
    instances = 0;
    preallocations = 0;
  }
  void invalidate() {
    instances = ~0U;
    preallocations = ~0U;
  }
  jobt_metrics& operator+=(const jobt_metrics& o) {
    instances += o.instances;
    preallocations += o.preallocations;
    return *this;
  }
  bool fills(const jobt_limits& o) const {
    return instances+preallocations >= o.max_instances;
  }
};

extern address_t local_address;

//
// Initialization routine
//
void initialize_cluster();

//
// This must be called regularly to expire pre-allocated job slots etc.
//
void maintain_cluster();

//
// Considering all host metrics, suggest the best host for job
// type allocation. Returns true if best host found, false if not possible
//
bool suggest_job_allocation(const std::string&, address_t&);

//
// Query routine for local allocation requests
//
bool can_allocate(const std::string&);

//
// See if serial number is already in use from that host
// (duplicated job alloc request)
//
bool is_serial_in_use(unsigned, const address_t&);

//
// Routines for recording a serial number that's being put
// into use, and for getting back the old packet.
//
void record_serial_in_use(unsigned, const address_t&, const char *, unsigned);
unsigned fillin_serial_host_reply(unsigned, const address_t&, char*);

//
// Registers local allocation and returns serial# of job slot
//
unsigned register_allocation(const std::string&, const std::string&, const address_t&);

//
// Update metrics for a host
//
void update_metrics(const address_t&, const std::string&, const jobt_metrics&);

//
// Invalidate a host (for unsubscription)
//
void invalidate_host_metrics(const address_t&);

//
// Update limits for a host
//
void update_limits(const address_t&, const std::string&, const jobt_limits&);

//
// Request our metrics
//
void get_local_metrics(std::map<std::string,jobt_metrics>&);

//
// Request our limits
//
void get_local_limits(std::map<std::string,jobt_limits>&);

//
// This is called when we subscribe to new hosts. We should
// set up a new limits/metrics record for the host if one
// doesn't exist already
//
void initialize_load_data_for_host(const address_t&);

//
// Debug stuff
//
void debug_print_load_data();

//
// This is used for validation of process spawn requests,
// we see if the host given is authorized to spawn a process
// of given type with given seq#
//
bool is_jobslot_valid(const address_t&, const std::string&, unsigned);

//
// We provide others with a way of hinting us to update the
// local metrics.  The argument is ``force''
//
void update_local_metrics(bool);

//
// When spawning a job, we need to know the username from
// the sequence number
//
std::string get_jobslot_username(unsigned);

//
// We tell the cluster management to drop a pre-allocated job-slot
// of some type whenever we spawn a process.
//
void drop_prealloc_slot(const std::string&);

#endif
