//
// Cluster handling -- this is the entire state machine
// for the message passing part of antsd
//
// $Id: jobnet.cc,v 1.18 2001/03/12 01:10:19 jakob Exp $
//
//  Copyright (C) 2000 Jakob Oestergaard <jakob@unthought.net>
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later
//  version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include "jobnet.hh"
#include "antsd.hh"
#include "cluster.hh"
#include "sock.hh"
#include "processes.hh"

#include "protocol.h"

#include <netdb.h>
#include <string.h>
#include <errno.h>

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#include <fstream>
#include <map>
#include <set>

static void send_subscription_request(const address_t&);

static void recv_subscribe(const char*, unsigned, const address_t&);
static void recv_unsubscribe(const char*, unsigned, const address_t&);
static void recv_job_alloc(const char*, unsigned, const address_t&);
static void recv_new_metric(const char*, unsigned, const address_t&);
static void recv_acknack(const char*, unsigned, const address_t&);
static void recv_req_new_metric(const char*, unsigned, const address_t&);
static void recv_new_limit(const char*, unsigned, const address_t&);
static void recv_req_new_limit(const char*, unsigned, const address_t&);

int JOBNET_PORT = 0;
int IDENT_PORT = 0;
int TCP_PROTO_NUM = 0;

const unsigned MAX_MSG = 8;
const unsigned ID_SUBSCRIBE = 0;
const unsigned ID_UNSUBSCRIBE = 1;
const unsigned ID_JOB_ALLOC = 2;
const unsigned ID_NEW_METRIC = 3;
const unsigned ID_ACKNACK = 4;
const unsigned ID_REQ_NEW_METRIC = 5;
const unsigned ID_NEW_LIMIT = 6;
const unsigned ID_REQ_NEW_LIMIT = 7;

void (*msghash[MAX_MSG])(const char*, unsigned, const address_t&) = {
  recv_subscribe,
  recv_unsubscribe,
  recv_job_alloc,
  recv_new_metric,
  recv_acknack,
  recv_req_new_metric,
  recv_new_limit,
  recv_req_new_limit
};

struct jobnet_header_t {
  unsigned magic;
  unsigned version;
  unsigned message;
  unsigned padding[5];
  bool is_valid() const { 
    return (magic == JOBNET_MAGIC) && (version == 0) && (message < MAX_MSG); 
  }
  void setup(unsigned m) {
    magic = JOBNET_MAGIC;
    version = 0;
    message = m;
    memset(padding, 0, sizeof(padding));
  }
};

struct jobnet_subscribe_t {
  char alias[ALIAS_LEN];
  bool i_know_you;
  char padding[256 - ALIAS_LEN - sizeof(bool)];
  bool is_valid() const { return !!alias[0]; }
  std::string parse_alias() const { 
    unsigned len = 0;
    while(len != ALIAS_LEN && alias[len]) len++;
    return std::string(alias, len);
  }
  void setup(const std::string& a, bool know) {
    if(a.size() >= ALIAS_LEN)
      throw internalerror("Too long alias in jobnet_subscribe setup");
    strcpy(alias, a.c_str());
    i_know_you = know;
  }
};


struct jobnet_unsubscribe_t {
  char alias[ALIAS_LEN];
  unsigned sequence;
  char padding[256 - sizeof(unsigned) - ALIAS_LEN];
  bool is_valid() const { return !!alias[0]; }
  std::string parse_alias() const { 
    unsigned len = 0;
    while(len != ALIAS_LEN && alias[len]) len++;
    return std::string(alias, len);
  }
};

struct jobnet_alloc_t {
  char type[TYPE_LEN];
  unsigned sequence;
  char username[USERNAME_LEN];
  char padding[256 - sizeof(unsigned) - TYPE_LEN - USERNAME_LEN];
  bool is_valid() const { return type[0] && username[0]; }
  std::string parse_type() const { 
    unsigned len = 0;
    while(len != TYPE_LEN && type[len]) len++;
    return std::string(type, len);
  }
  std::string parse_username() const { 
    unsigned len = 0;
    while(len != USERNAME_LEN && username[len]) len++;
    return std::string(username, len);
  }
  void setup(const std::string& t, const std::string& u, unsigned s) {
    if(t.size() >= TYPE_LEN)
      throw internalerror("Too long type in jobnet_alloc packet setup");
    if(u.size() >= USERNAME_LEN)
      throw internalerror("Too long username in jobnet_alloc packet setup");
    strcpy(type, t.c_str());
    strcpy(username, u.c_str());
    sequence = s;
  }    
};

struct jobnet_metric_t {
  char type[TYPE_LEN];
  unsigned instances;
  unsigned preallocations;
  char padding[256 - TYPE_LEN - 2*sizeof(unsigned)];
  bool is_valid() const { return !!type[0]; }
  void assign(const jobt_metrics& m) {
    instances = m.instances;
    preallocations = m.preallocations;
  }
  std::string parse_type() const { 
    unsigned len = 0;
    while(len != TYPE_LEN && type[len]) len++;
    return std::string(type, len);
  }
  operator jobt_metrics() const {
    jobt_metrics m;
    m.instances = instances;
    m.preallocations = preallocations;
    return m;
  }
};

struct jobnet_acknack_t {
  unsigned ack;
  unsigned sequence;
  unsigned reason;
  unsigned extra;
  unsigned padding[4];
  bool is_valid() const { return ((ack&1)==ack); }
  void setup(unsigned a, unsigned s, unsigned r = 0, unsigned e = 0) {
    ack = a;
    sequence = s;
    reason = r;
    extra = e;
  }
};

struct jobnet_limit_t {
  char type[TYPE_LEN];
  unsigned max_instances;
  unsigned preferred_instances;
  char padding[256 - TYPE_LEN - 2*sizeof(unsigned)];
  bool is_valid() const { return !!type[0]; }
  void assign(const jobt_limits& m) {
    max_instances = m.max_instances;
    preferred_instances = m.preferred_instances;
  }
  std::string parse_type() const { 
    unsigned len = 0;
    while(len != TYPE_LEN && type[len]) len++;
    return std::string(type, len);
  }
  operator jobt_limits() const {
    jobt_limits m;
    m.max_instances = max_instances;
    m.preferred_instances = preferred_instances;
    return m;
  }
};

int jobsock = -1;
typedef std::map<address_t, std::string> jobnet_addresses_t;
static jobnet_addresses_t jobnet_addresses;
static std::set<unsigned> hostfile_addresses;
static std::map<address_t, time_t> jobnet_last_message;
static std::map<address_t, time_t> jobnet_last_limits_update;
static char * replybuf = 0;
static unsigned sequencenr = 0;

struct callback_t {
  void(*f)(unsigned, unsigned, unsigned, unsigned, const address_t&);
  time_t ctime;
};
std::map<unsigned, callback_t> callbacks;

static unsigned localhost_address = 0;

struct retransmit_t {
  char * buf;
  unsigned len;
  address_t host;
};

std::map<unsigned, retransmit_t> retransmits;

//
// We need a simple checksum routine
//
static unsigned short checksum(unsigned len, const char *pkg)
{
  unsigned nleft = len;
  const unsigned short* addr = reinterpret_cast<const unsigned short*>(pkg);
  const unsigned short* w = addr;
  int sum = 0;
  unsigned short answer = 0;

  //
  // 32 bit accumulator gets 16 bit words added
  //
  while (nleft > 1)  {
    sum += *w++;
    nleft -= 2;
  }

  //
  // Need one last byte, maybe
  //
  if (nleft == 1) {
    *(unsigned char *)(&answer) = *(unsigned char *)w ;
    sum += answer;
  }

  //
  // Upper 16 bit (carry bits from low 16 word) are folded
  // back onto the lower 16 bits
  //
  sum = (sum >> 16) + (sum & 0xffff);  // add upper and lower 16 bit word
  sum += (sum >> 16);                  // add upper 16 bit (carry) from prev. fold
  answer = ~sum;
  return answer;
}

static bool check_packet(unsigned len, const char* pkt)
{
  if(len < 2) throw internalerror("check_packet was given a short packet");
  int csum = checksum(len - 2, pkt);
  const unsigned short* pktsum = reinterpret_cast<const unsigned short*>(pkt + len - 2);
  if(csum != *pktsum) {
    log_info("Checksum mismatch", 1);
    return false;
  }
  return true;
}

static void send_packet(unsigned len, const address_t& to)
{
  //
  // Calc checksum and append
  //
  if(len + 2 >= MAX_PACKET) {
    log_info("send_packet() got oversized packet", 1);
    return;
  }
  u_short * sig = (u_short*)(replybuf + len);
  *sig = checksum(len, replybuf);
  len += 2;

  jobnet_header_t * validator = reinterpret_cast<jobnet_header_t*>(replybuf);
  if(!validator->is_valid()) {
#if 0
    std::cerr << "Sending packet with: "
	      << " l: " << len
	      << " m: " << *reinterpret_cast<unsigned*>(replybuf)
	      << " v: " << *reinterpret_cast<unsigned*>(replybuf + sizeof(unsigned))
	      << " g: " << *reinterpret_cast<unsigned*>(replybuf + 2*sizeof(unsigned))
	      << std::endl;
    abort();
#endif
    throw internalerror("Would have sent bad packet!");
  }

  const struct sockaddr_in sto(to);
  int rc = sendto(jobsock, replybuf, len, 0, (struct sockaddr*)&sto, sizeof(sto));
  if(rc != (signed)len) {
    log_info("sendto() failed.", 1);
    return;
  }
}


void initialize_jobnet()
{
  //
  // Setup numbers needed later on
  //
  JOBNET_PORT = htons(JOBNET_PORT_DECIMAL);
  IDENT_PORT = htons(IDENT_PORT_DECIMAL);

  //
  // Read the hostfile
  //
  log_info("Resolving hosts...");
  ifstream hostfile(HOSTFILE.c_str());
  while(hostfile) {
    std::string name;
    hostfile >> name;
    struct hostent * hent = gethostbyname(name.c_str());
    if(!hent) continue;
    if(hent->h_length != sizeof(unsigned)) {
      log_info("Address length mismatch!", 1);
      return;
    }
    for(char** addr = hent->h_addr_list; *addr; addr++) {     
      unsigned * a = reinterpret_cast<unsigned*>(*addr);
      hostfile_addresses.insert(*a);
    }
  }

  if(!hostfile_addresses.size())
    exit_fatal("There are no hosts in " + HOSTFILE + "... exiting.");

  //
  // Resolve "localhost"
  //
  struct hostent * hent = gethostbyname("localhost");
  if(!hent) {
    log_info("Could not resolve \"localhost\"", 1);
    return;
  }
  if(hent->h_length != sizeof(unsigned)) {
    log_info("Address length mismatch!", 1);
    return;
  }
  localhost_address = *reinterpret_cast<unsigned*>(hent->h_addr_list[0]);

  //
  // Alloc buffers
  //
  replybuf = new char[MAX_PACKET];

  log_info("JobNet Initialization Complete.");
}


void setup_jobnet()
{
  jobsock = socket(AF_INET, SOCK_DGRAM, 0);
  if(jobsock == -1) {
    log_info("Sorry, cannot create socket for jobnet", 1);
    return;
  }

  struct sockaddr_in myaddr;
  memset(&myaddr, 0, sizeof(myaddr));
  myaddr.sin_family = AF_INET;
  myaddr.sin_port = JOBNET_PORT;
  unsigned adr = INADDR_ANY;
  memcpy(&myaddr.sin_addr, &adr, sizeof(adr));

  int rc = bind(jobsock, (const struct sockaddr*)&myaddr, sizeof(struct sockaddr_in));
  if(rc) {
    log_info("Sorry, cannot bind to JobNet port - maybe antsd is already running?", 1);
    close(jobsock);
    jobsock = -1;
  } else {
    log_info("Bound to JobNet socket - ok");
  }
}


void read_jobnet()
{
  static char * buf = 0;
  if(!buf)
    buf = new char[MAX_PACKET];

  struct sockaddr_in from;
  unsigned fromlen = sizeof(from);
  int rc = recvfrom(jobsock, buf, MAX_PACKET, 0, reinterpret_cast<struct sockaddr*>(&from), &fromlen);

  if(rc == -1) {
    log_info("Short read from JobNet socket! (" + std::string(strerror(errno)) + ")", 1);
    return;
  }
  if(!check_packet(rc, buf)) {
    log_info("Packet checksum mismatch!", 1);
    return;
  }

  jobnet_header_t * head = reinterpret_cast<jobnet_header_t*>(buf);
  if(! head->is_valid()) {
    log_info("Bad packet header", 1);
    return;
  }

  msghash[head->message](buf + sizeof(jobnet_header_t), rc - sizeof(jobnet_header_t) - 2, from);
}




static void recv_subscribe(const char* pkt, unsigned len, const address_t& from)
{
  if(len < sizeof(jobnet_subscribe_t)) {
    log_info("Undersized subscription packet received from " + from.toString(), 1);
    return;
  }

  const jobnet_subscribe_t * head = reinterpret_cast<const jobnet_subscribe_t*>(pkt);
  if(! head->is_valid()) return;

  //
  // See if we can find this host in /etc/hosts.equiv
  //
  std::set<unsigned>::const_iterator hi = hostfile_addresses.find(from.inet_addr);
  if(hi == hostfile_addresses.end()) {
    log_info("Got valid subscription from a host ("  + from.toString() + ") not in the hostfile", 1);
    return;
  }

  const string alias = head->parse_alias();

  log_info("Got valid and authorized subscription from \"" + alias + "\" (" + from.toString() + ")");

  //
  // Tell the cluster management to set up a limits/metrics record
  // for this host  -  if the host is alreay registered, it may be
  // because it was shut down and re-started, and in that case it
  // could have been re-configured, and we also want to initialize
  // the load data.  Therefore this line is called before the check
  // for previous registrations...
  //
  initialize_load_data_for_host(from);

  //
  // If it's already subscribed, update or ignore.
  //
  jobnet_addresses_t::iterator existing = jobnet_addresses.find(from);
  if(existing != jobnet_addresses.end()) {
    if(existing->second != alias) {
      log_info("Address \"" +  from.toString() + "\" is already subscribed, will change alias from \""
	       + existing->second
               + "\" to \"" + alias + "\"", 1);
      existing->second = alias;
      return;
    } else {
      if(head->i_know_you) {
        log_info("Address \"" +  from.toString()
		 + "\" is already subscribed under same alias and host knows us already.", 1);
        return;
      }
      log_info("Address\"" +  from.toString()
	       + "\" is already subscribed, but host doesn't know us. Sending subscription.", 1);
      send_subscription_request(from);
      return;
    }
  }

  //
  // New subscriptions are entered.
  //
  jobnet_addresses[from] = alias;

  log_info("New host \"" + alias + "\" \"" +  from.toString() + "\"is now subscribed");

  //
  // Send a subscription request in response, if this was a new
  // subscription
  //
  time_t now = time(0);

  std::map<address_t, time_t>::const_iterator hosti = jobnet_last_message.find(from);
  if(hosti == jobnet_last_message.end()
     || hosti->second + JOBNET_RESUBSCRIBE_AFTER < now) 
    send_subscription_request(from);

  //
  // Prevent immediate timeout
  //
  jobnet_last_message[from] = now;
}


static void recv_unsubscribe(const char* pkt, unsigned len, const address_t& from)
{
  if(len < sizeof(jobnet_unsubscribe_t)) {
    log_info("Undersized unsubscription packet received from \"" +  from.toString() + "\"", 1);
    return;
  }

  const jobnet_unsubscribe_t * head = reinterpret_cast<const jobnet_unsubscribe_t*>(pkt);
  if(! head->is_valid()) return;

  //
  // Make sure that alias and source port matches the subscription
  //
  jobnet_addresses_t::iterator hi = jobnet_addresses.find(from);
  if(hi == jobnet_addresses.end()) {
    log_info("Received unsubscription from unknown host:\"" +  from.toString() + "\"", 1);
    return;
  }
  if(hi->second != head->parse_alias()) {
    log_info("Received unsubscription from subscribed host\"" +  from.toString()
	     + "\", but alias mismatched. Ignoring message", 1);
    return;
  }

  log_info("Will unsubscribe host \"" + hi->second + "\" as requested by \"" +  from.toString() + "\"");
  jobnet_addresses.erase(hi);  
}


static void recv_job_alloc(const char* pkt, unsigned len, const address_t& from)
{
  if(len < sizeof(jobnet_alloc_t)) {
    log_info("Undersized allocation packet received from \"" +  from.toString() + "\"", 1);
    return;
  }
  const jobnet_alloc_t * head = reinterpret_cast<const jobnet_alloc_t*>(pkt);
  if(! head->is_valid()) return;

  jobnet_header_t * rhead = reinterpret_cast<jobnet_header_t*>(replybuf);
  jobnet_acknack_t * racknack = reinterpret_cast<jobnet_acknack_t*>(replybuf + sizeof(jobnet_header_t));
  rhead->setup(ID_ACKNACK);

  //
  // See that host is known and subscribed
  //
  jobnet_addresses_t::iterator hi = jobnet_addresses.find(from);
  if(hi == jobnet_addresses.end()) {
    //
    // If host is known, we send it a NACK telling it it's unsubscribed
    //
    std::set<unsigned>::const_iterator hfi = hostfile_addresses.find(from.inet_addr);
    if(hfi == hostfile_addresses.end()) {
      log_info("Received job allocation request from unsubscribed unknown host \"" +  from.toString() + "\"", 1);
      return;
    }

    racknack->setup(0, head->sequence, 1); // reson = 1  =>  you're not subscribed
    send_packet(sizeof(jobnet_header_t) + sizeof(jobnet_acknack_t), from);
      
    return;
  }

  //
  // If that host already requested something using the
  // same sequence number, we should just retransmit our
  // response
  //
  if(is_serial_in_use(head->sequence, from)) {
    log_info("We got a secondary allocation request from \"" +  from.toString() + "\"... Resending response", 1);
    std::cerr << "Sequence is " << head->sequence << std::endl;
    
    unsigned len = fillin_serial_host_reply(head->sequence, from, replybuf);
    send_packet(len, from);

    return;
  }

  //
  // See if we can allocate the job slot
  //
  const string jobtype = head->parse_type();
  if(!can_allocate(jobtype)) {
    log_info("Full load: Refusing job allocation of type \"" + jobtype
	     + "\" for remote node \"" +  from.toString() + "\"", 1);

    racknack->setup(0, head->sequence, 0); // reson = 0  =>  we're full
    send_packet(sizeof(jobnet_header_t) + sizeof(jobnet_acknack_t), from);
    
    return;
  }
  const string username = head->parse_username();

  const unsigned serial = register_allocation(jobtype, username, from);
  if(!serial) {
    log_info("register_allocation() with \"" +  from.toString() + "\" failed. Aborting transaction.", 1);
    return;
  }

  log_info("Allocating job slot of type \"" + jobtype + "\" for user \"" 
	   + username + "\" on remote node (\"" +  from.toString() + "\")");


  unsigned bound_port = startup_subprocess(from, serial, username, jobtype);

  if(!bound_port) {
    // Duh !
    racknack->setup(0, head->sequence, 0); // reson = 0  =>  we're full
    send_packet(sizeof(jobnet_header_t) + sizeof(jobnet_acknack_t), from);
    return;
  }

  //
  // Send back positive reply
  //
  racknack->setup(1, head->sequence, serial, bound_port); // we use the reason field for serial#

  record_serial_in_use(head->sequence, from, replybuf,
		       sizeof(jobnet_header_t) + sizeof(jobnet_acknack_t));

  send_packet(sizeof(jobnet_header_t) + sizeof(jobnet_acknack_t), from);

  send_metrics(from); // must be called after register_allocation which called update_local_metrics(false)

  log_info("* Completed recv_job_alloc() with \"" +  from.toString() + "\"");
}


static void recv_new_metric(const char* pkt, unsigned len, const address_t& from)
{
  if(len < sizeof(jobnet_metric_t)) {
    log_info("Undersized metric packet received from \"" +  from.toString() + "\"", 1);
    return;
  }
  const jobnet_metric_t * head = reinterpret_cast<const jobnet_metric_t*>(pkt);
  if(! head->is_valid()) return;

  jobnet_addresses_t::iterator hi = jobnet_addresses.find(from);
  if(hi == jobnet_addresses.end()) {
    log_info("Received new metric from host not in cluster: \"" +  from.toString() + "\"", 1);
    return;
  }

  log_info("Validated metric update request from peer \""
	   +  from.toString() + "\" for jobtype " + head->parse_type());

  update_metrics(from, head->parse_type(), *head);

  //
  // Remember this moment for re-subscription heuristics
  //
  jobnet_last_message[from] = time(0);
  
}

static void recv_acknack(const char* pkt, unsigned len, const address_t& from)
{
  if(len < sizeof(jobnet_acknack_t)) {
    log_info("Undersized acknack packet received from \"" +  from.toString() + "\"", 1);
    return;
  }
  const jobnet_acknack_t * head = reinterpret_cast<const jobnet_acknack_t*>(pkt);
  if(! head->is_valid()) {
    log_info("Bad header in recv_acknack packet received from \"" +  from.toString() + "\"", 1);
    return;
  }

  //
  // See that host is valid
  //
  jobnet_addresses_t::iterator hi = jobnet_addresses.find(from);
  if(hi == jobnet_addresses.end()) {
    log_info("Received acknack from unknown host \"" +  from.toString() + "\"", 1);
    return;
  }

  //
  // Lookup sequence number
  //
  std::map<unsigned, callback_t>::iterator ci = callbacks.find(head->sequence);
  if(ci == callbacks.end()) {
    log_info("Received ack/nack for unknown sequence number \"" +  from.toString() + "\"", 1);
    std::cerr << "unknown seq is " << head->sequence << std::endl;
    return;
  }

  log_info("Received good acknack from \"" +  from.toString() + "\", will treat.");
  ci->second.f(head->ack, head->sequence, head->reason, head->extra, from);
  callbacks.erase(ci);
}

static void recv_req_new_metric(const char* pkt, unsigned len, const address_t& from)
{
  //
  // There's no information here... Just check that host is known and send
  // back metrics
  //
  jobnet_addresses_t::iterator hi = jobnet_addresses.find(from);
  if(hi == jobnet_addresses.end()) {
    log_info("Received new_metric request from unknown host \"" +  from.toString() + "\"", 1);
    return;
  }
  if(from.inet_addr == local_address.inet_addr) {
    log_info("Detected metric feedback.. Bug or attack ?", 1);
    return;
  }

  send_metrics(from);  
}

static void recv_new_limit(const char* pkt, unsigned len, const address_t& from)
{
  if(len < sizeof(jobnet_limit_t)) {
    log_info("Undersized limit packet received from \"" +  from.toString() + "\"", 1);
    return;
  }
  const jobnet_limit_t * head = reinterpret_cast<const jobnet_limit_t*>(pkt);
  if(! head->is_valid()) return;

  jobnet_addresses_t::iterator hi = jobnet_addresses.find(from);
  if(hi == jobnet_addresses.end()) {
    log_info("Received new limit from host (\"" +  from.toString()
	     + "\") not in cluster", 1);
    return;
  }

  log_info("Validated limit update request from peer \"" +  from.toString() + "\"");

  update_limits(from, head->parse_type(), *head);  

  jobnet_last_limits_update[from] = time(0);
  
}

static void recv_req_new_limit(const char* pkt, unsigned len, const address_t& from)
{
  //
  // There's no information here... Just check that host is known and send
  // back limits
  //
  jobnet_addresses_t::iterator hi = jobnet_addresses.find(from);
  if(hi == jobnet_addresses.end()) {
    log_info("Received new_limit request from unknown host \"" +  from.toString() + "\"", 1);
    return;
  }

  send_limits(from);  
}

void except_jobnet()
{
  log_info("Caught exception on JobNet socket. Closing it down", 1);
  close(jobsock);
  jobsock = -1;
}

void shutdown_jobnet()
{
  log_info("JobNet connection shut down.", 1);
  close(jobsock);
  jobsock = -1;
}

static void send_subscription_request(const address_t& host)
{
  jobnet_header_t * phead = reinterpret_cast<jobnet_header_t*>(replybuf);
  jobnet_subscribe_t * psubscr = reinterpret_cast<jobnet_subscribe_t*>(replybuf + sizeof(jobnet_header_t));
  phead->setup(ID_SUBSCRIBE);
  char myname[ALIAS_LEN];
  if(gethostname(myname, ALIAS_LEN)) {
    log_info("Failed retrieval of my own hostname... Using \"antsd\" instead", 1);
    strcpy(myname, "antsd");
  }
  // See if we know that fellow already
  jobnet_addresses_t::const_iterator ji = jobnet_addresses.find(host);
  psubscr->setup(myname, ji != jobnet_addresses.end());
  send_packet(sizeof(jobnet_header_t) + sizeof(jobnet_subscribe_t), host);
  log_info("Sent subscription request to remote host \"" +  host.toString() + "\"");
}

void send_metrics_to_all()
{
 for(jobnet_addresses_t::const_iterator hi = jobnet_addresses.begin();
      hi != jobnet_addresses.end(); hi++)
   if(hi->first != local_address) {
      send_metrics(hi->first);
    }
}

void send_limits_to_all()
{
  const time_t now = time(0);
  for(jobnet_addresses_t::const_iterator hi = jobnet_addresses.begin();
      hi != jobnet_addresses.end(); hi++)
    if(hi->first != local_address) {
      std::map<address_t, time_t>::iterator limi = jobnet_last_limits_update.find(hi->first);
      if(limi == jobnet_last_limits_update.end()) {
        send_limits(hi->first);
        jobnet_last_limits_update[hi->first] = 0;
      } else {
        if(limi->second + JOBNET_UPDATE_LIMITS_INTERVAL < now) {
          send_limits(hi->first);
        }
      }
    }
}


void maintain_jobnet()
{
  const time_t now = time(0);

  debug_print_load_data();

  //
  // Expire callbacks, 
  //
  for(std::map<unsigned, callback_t>::iterator ci = callbacks.begin();
      ci != callbacks.end(); ) {
    std::map<unsigned, callback_t>::iterator next = ci; next++;
    if(ci->second.ctime + JOBNET_CALLBACK_MAXAGE < now) {
      log_info("Expired callback", 1);
      callbacks.erase(ci);
    }
    ci = next;
  }

  //
  // Send out new metrics and limits where needed
  //
  send_limits_to_all();
  send_metrics_to_all();
  
  //
  // Re-send subscription requests to hosts in the cluster once in a while
  // We send to any host we haven't heard from in some time...
  //
  std::set<address_t> unsubscribe_hosts;
  for(std::set<unsigned>::const_iterator hi = hostfile_addresses.begin();
      hi != hostfile_addresses.end(); hi++) {
    address_t host;
    host.inet_addr = *hi;
    host.inet_port = JOBNET_PORT;
    if(host == local_address) continue;
    jobnet_addresses_t::const_iterator ji = jobnet_addresses.find(host);
    if(ji != jobnet_addresses.end()) {
      //
      // Check time
      //
      std::map<address_t, time_t>::const_iterator ti = jobnet_last_message.find(host);
      if(ti != jobnet_last_message.end()
         && ti->second + JOBNET_RESUBSCRIBE_AFTER > now) continue;

      //
      // If we haven't heard from host in some time, unregister it
      //
      if(ti->second + JOBNET_UNSUBSCRIBE_AFTER < now) {
        log_info("Unsubscribing host alias \"" + ji->second + "\" because of timeout", 1);
        unsubscribe_hosts.insert(host);
        continue;
      }

      //
      // Resubscribe because we haven't heard from this host in a while
      //
      send_subscription_request(host);

      continue;
    }
    //
    // Send subscription request to host because it haven't registered with us
    //
    send_subscription_request(host);
  }

  //
  // Perform the actual unsubscription
  //
  while(unsubscribe_hosts.begin() != unsubscribe_hosts.end()) {
    jobnet_addresses.erase(*unsubscribe_hosts.begin());
    invalidate_host_metrics(*unsubscribe_hosts.begin());
    //
    unsubscribe_hosts.erase(unsubscribe_hosts.begin());
  }

}

void register_acknack_callback(unsigned seq, void(*f)(unsigned, unsigned, unsigned, unsigned, const address_t&))
{
  if(callbacks.find(seq) != callbacks.end())
    log_info("Callback sequence number conflict. Ignoring.... Inserting new callback.", 1);

  callback_t nc;
  nc.f = f;
  nc.ctime = time(0);
  callbacks.insert(make_pair(seq, nc));
}


static bool remote_allocation_recall(int retries, unsigned alloc_sequence)
{
  std::map<unsigned,retransmit_t>::iterator ri = retransmits.find(alloc_sequence);
  if(ri == retransmits.end()) {
    log_info("Abandoning re-transmit of allocation request that doesn't exist", 1);
    return false;
  }

  if(!retries) {
    log_info("Re-transmit allocation requests failed!", 1);
    // Tell client about it
    notify_client_alloc_failure(ri->second.host, alloc_sequence);
    // Clean up
    delete[] ri->second.buf;
    retransmits.erase(ri);
    return false;
  }

  //
  // Sanity check - did the host disappear ?
  //
  jobnet_addresses_t::const_iterator ji = jobnet_addresses.find(ri->second.host);
  if(ji == jobnet_addresses.end()) {
    log_info("Abandoning re-transmit of allocation request to host not in jobnet", 1);
    delete[] ri->second.buf;
    retransmits.erase(ri);
    return false;
  }

  memcpy(replybuf, ri->second.buf, ri->second.len);
  send_packet(ri->second.len, ri->second.host);

  if(retries < (signed)JOBNET_ALLOC_RETRIES-1) {
    log_info("Re-sending remote allocation request", 1);
    std::cerr << "serial number is " << alloc_sequence << std::endl;
  }

  return true;
}

unsigned request_remote_allocation(const std::string& type, const std::string& username, const address_t& host)
{
  //
  // We simply register the allocation request in the
  // retry queue.
  //
  //
  // Just send a plain job slot allocation request to remote node
  //
  jobnet_header_t * phead = reinterpret_cast<jobnet_header_t*>(replybuf);
  jobnet_alloc_t * palloc = reinterpret_cast<jobnet_alloc_t*>(replybuf + sizeof(jobnet_header_t));

  phead->setup(ID_JOB_ALLOC);
  palloc->setup(type, username, ++sequencenr);

  retransmit_t retr;
  retr.len = sizeof(jobnet_header_t) + sizeof(jobnet_alloc_t);
  retr.buf = new char[retr.len];
  memcpy(retr.buf, replybuf, retr.len);
  retr.host = host;

  std::map<unsigned, retransmit_t>::iterator rti = retransmits.find(sequencenr);
  if(rti != retransmits.end()) {
    log_info("This is odd - we had a sequence number conflict ! - We will leak memory.", 1);
    retransmits.erase(rti);
  }
  retransmits.insert(std::make_pair(sequencenr,retr));

  add_recall_routine(remote_allocation_recall, JOBNET_ALLOC_RETRY_DELAY, 
		     JOBNET_ALLOC_RETRIES, sequencenr);

  return sequencenr;
}

void acknowledge_remote_allocation(unsigned sequence, const address_t& host)
{
  //
  // Tell the main antsd to forget about this recall method
  //
  remove_recall_routine(remote_allocation_recall, sequence);

  //
  // Just kill the retransmit handler.
  //
  std::map<unsigned,retransmit_t>::iterator ri = retransmits.find(sequence);
  if(ri == retransmits.end()) {
    //
    // Luck is nothing but an absence of bad luck.  This case _can_ happen
    // of course, but if this line occurs often in your logs, something
    // other than luck is playing in.
    //
    std::cerr << "* Nothing with sequence " << sequence << " in retransmit map"
	      << std::endl;
    log_info("When acknowledging remote allocation: retransmits timed out, we were lucky", 1);
    return;
  }

  delete[] ri->second.buf;
  retransmits.erase(ri);
}

void request_new_metrics(const address_t& host)
{
  jobnet_header_t * phead = reinterpret_cast<jobnet_header_t*>(replybuf);

  phead->setup(ID_REQ_NEW_METRIC);
  send_packet(sizeof(jobnet_header_t), host);

  log_info("Sent request for new metrics to remote host \"" +  host.toString() + "\"");
}

void send_metrics(const address_t& host)
{
  //
  // Send local metrics to host
  //
  std::map<std::string,jobt_metrics> metrics;
  get_local_metrics(metrics);

  for(std::map<std::string,jobt_metrics>::const_iterator mi = metrics.begin();
      mi != metrics.end(); mi++) {
    jobnet_header_t * phead = reinterpret_cast<jobnet_header_t*>(replybuf);
    jobnet_metric_t * pmetric = reinterpret_cast<jobnet_metric_t*>(replybuf + sizeof(jobnet_header_t));

    phead->setup(ID_NEW_METRIC);
    
    strcpy(pmetric->type, mi->first.c_str());
    pmetric->assign(mi->second);

    send_packet(sizeof(jobnet_header_t) + sizeof(jobnet_metric_t), host);
  }
  
  log_info("Sent new metrics to remote host \"" +  host.toString() + "\"");
}

void request_new_limits(const address_t& host)
{
  jobnet_header_t * phead = reinterpret_cast<jobnet_header_t*>(replybuf);

  phead->setup(ID_REQ_NEW_LIMIT);
  send_packet(sizeof(jobnet_header_t), host);

  log_info("Sent request for new limits to remote host \"" +  host.toString() + "\"");
}

void send_limits(const address_t& host)
{
  //
  // Send local limits to host
  //
  std::map<std::string,jobt_limits> limits;
  get_local_limits(limits);

  for(std::map<std::string,jobt_limits>::const_iterator mi = limits.begin();
      mi != limits.end(); mi++) {
    jobnet_header_t * phead = reinterpret_cast<jobnet_header_t*>(replybuf);
    jobnet_limit_t * plimit = reinterpret_cast<jobnet_limit_t*>(replybuf + sizeof(jobnet_header_t));

    phead->setup(ID_NEW_LIMIT);
    
    strcpy(plimit->type, mi->first.c_str());
    plimit->assign(mi->second);

    send_packet(sizeof(jobnet_header_t) + sizeof(jobnet_limit_t), host);
  }
  
  log_info("Sent new limits to remote host \"" +  host.toString() + "\"");
}


bool is_host_in_jobnet(address_t& host)
{
  //
  // Special case for localhost
  //
  if(host.inet_addr == localhost_address)
    host.inet_addr = local_address.inet_addr;
  //
  // Since we do not subscribe to ourselves, handle this matter as well
  //
  if(host.inet_addr == local_address.inet_addr)
    return true;
  //
  // Look at rest of addresses
  //
  for(std::map<address_t, std::string>::const_iterator jni = jobnet_addresses.begin();
      jni != jobnet_addresses.end(); jni++)
    if(jni->first.inet_addr == host.inet_addr)
      return true;
  return false;
}

void request_metrics_from_all()
{
 for(jobnet_addresses_t::const_iterator hi = jobnet_addresses.begin();
      hi != jobnet_addresses.end(); hi++)
   if(hi->first != local_address) {
     request_new_metrics(hi->first);
   }
}
