//
// Definitions for the jobnet code
//
// $Id: jobnet.hh,v 1.10 2001/02/11 07:05:07 jakob Exp $
//
//  Copyright (C) 2000 Jakob Oestergaard <jakob@unthought.net>
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later
//  version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#ifndef __JOBNET_HH
#define __JOBNET_HH

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <string>

struct address_t { 
  unsigned inet_addr;
  int inet_port;
  address_t() {}
  address_t(const struct sockaddr_in& i)
    : inet_addr(i.sin_addr.s_addr), inet_port(i.sin_port) {}
  bool operator==(const address_t& o) const {
    return inet_addr == o.inet_addr
      && inet_port == o.inet_port;
  }
  operator struct sockaddr_in() const {
    struct sockaddr_in i;
    memset(&i, 0, sizeof(i));
    i.sin_family = AF_INET;
    i.sin_addr.s_addr = inet_addr;
    i.sin_port = inet_port;
    return i;
  }
  bool operator<(const address_t& o) const {
    if(inet_addr < o.inet_addr) return true;
    if(inet_addr > o.inet_addr) return false;
    return inet_port < o.inet_port;
  }
  std::string toString() const { 
    struct in_addr i;  i.s_addr = inet_addr;
    return std::string(inet_ntoa(i));
  }
};

extern int jobsock;


//
// Call-once. Buffer allocation, config reading etc.
//
void initialize_jobnet();

//
// Attempt jobnet setup.
//
void setup_jobnet();

//
// Read from the net, do what needs be done
//
void read_jobnet();

//
// Handle exceptions on the socket
//
void except_jobnet();

//
// Shutdown the JobNet socket
//
void shutdown_jobnet();

//
// Call this once in a while to tidy up the protocol stack and
// send out periodical metric updates to the net
//
void maintain_jobnet();

//
// Callbacks for ACK/NACKs with specific sequence numbers
// can be registered here
// Args are:  ack/nack, sequence, reason, from_address
void register_acknack_callback(unsigned, void(*)(unsigned, unsigned, unsigned, unsigned, const address_t&));

//
// Request job allocation on remote node. Returns the sequence number
// the ack / nack will hold.
//
unsigned request_remote_allocation(const std::string&, const std::string&, const address_t&);

//
// When a response to the request_remote_allocation() is received,
// checked, authenticated, validated, measured and weighed, we
// should acknowledge the ack/nack in order to kill any further retransmits.
//
// Args are:  sequence (used in alloc request) and peer address
//
void acknowledge_remote_allocation(unsigned, const address_t&);

//
// Request metric update from host
//
void request_new_metrics(const address_t&);

//
// Send our metrics to a given host
//
void send_metrics(const address_t&);

//
// Request limit update from host
//
void request_new_limits(const address_t&);

//
// Send our limits to a given host
//
void send_limits(const address_t&);

//
// Verify that host is known - port numbers may mismatch
// If the argument given is the localhost, 127.0.0.1 will
// be changed to the official jobnet IP of the local host.
bool is_host_in_jobnet(address_t&);

//
// Convenient routines when we know that status changes
//
void send_limits_to_all();
void send_metrics_to_all();
void request_metrics_from_all();


#endif
