//
// Process table implementation
//
// $Id: processes.cc,v 1.20 2001/03/12 01:10:19 jakob Exp $
//
//  Copyright (C) 2000 Jakob Oestergaard <jakob@unthought.net>
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later
//  version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include "processes.hh"
#include "antsd.hh"
#include "jobnet.hh"

#include <sys/resource.h>
#include <stdio.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <netinet/tcp.h>
#include <syslog.h>

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif

#include <fstream>
#include <algorithm>
#include <set>
#include <map>

std::map<int,process> process_table;
static char * buffer = 0;
int process_sock = -1;

jobt_metrics process::get_jobt_metrics() const
{
  jobt_metrics lim;

  //
  // /proc/[pid]/statm contains:
  // size resident shared trs drs lrs dt
  // We only care about size. And we only
  // use it for deciding whether the process
  // should be counted as existing or not.
  //
  char procfile[1024];
  sprintf(procfile, "/proc/%i/statm", pid);
  std::ifstream pfile(procfile);
  unsigned long memory;
  pfile >> memory;

  if(memory)
    lim.instances = 1;  
  else
    lim.instances = 0;  // Zombies count as not being there

  return lim;
}


static void clean_away_process_connection(int con_fd)
{
  std::map<int,process>::iterator pi = process_table.find(con_fd);
  if(pi == process_table.end()) {
    log_info("Most odd - we're unable to find the connection fd"
	     " in clean_away_process_connection()");
    return;
  }

  //
  // First see if process died
  //
  if(pi->second.pid != 0) {
    int status;
    int rc = waitpid(pi->second.pid, &status, WNOHANG);
    if(rc == pi->second.pid) {
      //
      // It died - great !
      //
      pi->second.pid = 0;

      //
      // Check stats !
      //
      if(WIFEXITED(status)) {
	if(WEXITSTATUS(status)) {
	  log_info("process handler exited with non-zero exit status!", 1);
	}
      } else if(WIFSIGNALED(status)) {
	char buf[512];
	sprintf(buf, "process handler terminated with signal %i", WTERMSIG(status));
	log_info(buf, 1);
      }

      log_info("Cleared a pid from a dead child");

    } else if(rc == 0) {
      //
      // It didn't die - well help it on it's way then !
      //
      if(pi->second.pid != 0)
	kill(pi->second.pid, SIGTERM);

      //
      // Return for now - let it die properly by itself,
      // we'll come back after next select() if it died.
      //
      log_info("Sent TERM signal to undead process handler", 1);

      return;
    } else if(rc == -1 && errno == ECHILD) {
      log_info("Odd - subprocess doesn't exist", 1);
    }
  }

  //
  // Close the pipe.
  //
  if(pi->first != -1)
    close(pi->first);

  //
  // Kill from table
  //
  process_table.erase(pi);

  log_info("Re-sending metrics to all, now that we know expired processes", 1);
  update_local_metrics(true);
  send_metrics_to_all();
}

void initialize_processes()
{
  buffer = new char[PROCESS_MESSAGE_LEN];
}


void maintain_processes()
{
  int stat;
  int rc;

  //
  // See if any died
  //
  while((rc = waitpid(-1, &stat, WNOHANG)) > 0) {
    //
    // Look up PID if it's ours.
    //
    std::map<int,process>::iterator pi = process_table.begin();
    while(pi != process_table.end())
      if(pi->second.pid == rc) break;
      else pi++;
    if(pi == process_table.end()) {
      log_info("A child died but we don't know about it...", 1);
      continue;
    }
    //
    // And terminate process.
    //
    clean_away_process_connection(pi->first);
  }

  time_t now = -1;

  //
  // See if any timed out
  //
  for(std::map<int,process>::iterator pi = process_table.begin();
      pi != process_table.end(); ) {
    if(pi->second.auth_ok) {
      pi++;
      continue;
    }

    //
    // By fetching the time in here, we save syscalls.
    //
    if(now == -1)
      now = time(0);

    if( pi->second.spawn_time + JOBCONN_AUTH_TIMEOUT <= now ) {
      //
      // It timed out - kill it.
      //
      log_info("Subprocess authentication timed out! Cleaning.");
      int fd = pi->first;
      pi++;
      clean_away_process_connection(fd);
      continue;
    }
    pi++;
    continue;
  }
}


void insert_process_pipe_set(fd_set& pset, int& max)
{
  for(std::map<int,process>::const_iterator pi = process_table.begin();
      pi != process_table.end(); pi++) {
    FD_SET(pi->first, &pset);
    max = std::max(max, pi->first);
  } 
}


void read_processes_pipe_set(const fd_set& pset)
{
  bool new_subjobs = false;
  //
  // In here, we read from the sub-processes we spawned
  // to handle job execution
  //
  std::set<int> expires;

  for(std::map<int,process>::iterator pi = process_table.begin();
      pi != process_table.end(); pi++)
    if(FD_ISSET(pi->first, &pset)) {
      unsigned msg;
      int rc = read(pi->first, &msg, sizeof(unsigned));

      if(rc != sizeof(unsigned)) {
	//
	// This happens when a process handler decides to die because it's
	// job completed.  It's perfectly normal.
	//
        log_info("job-handler is no more - cleaning remains...");
        expires.insert(pi->first);
        continue;
      }

      //
      // If the sub-process is writing to us, it means that the job
      // authenticated and has started.  The subprocess-handler will
      // *never* write anything else.
      //
      // Just check that the msg sent matches the expected value

      log_info("we received the message that auth in the child is ok");

      if(msg != subprocess_ok_msg) {
	log_info("sub-process sent us garbage - cleaning it.");
	expires.insert(pi->first);
	continue;
      }
      pi->second.auth_ok = true;

      //
      // As the job is now running, drop the pre-allocation
      //
      drop_prealloc_slot(pi->second.type);
      new_subjobs = true;
    }  

  //
  // Remove the process we found had died from the process_table.
  // Also re-send metrics if we did indeed change the process_table.
  //

  while(expires.begin() != expires.end()) {
    clean_away_process_connection(*expires.begin());
    expires.erase(expires.begin());
  }

  if(new_subjobs) {
    update_local_metrics(true);
    send_metrics_to_all();
  }
}


unsigned startup_subprocess(const address_t& from, unsigned serial, 
			    const std::string& username, const std::string& type)
{
  //
  // Bind to a random port, fork() sub-process handler, send back serial# and port#.
  //
  int subsock = socket(AF_INET, SOCK_STREAM, 0);
  if(subsock == -1) {
    // Shoot !  We're definitely out of resources now!
    log_info("Cannot allocate new socket!");
    return 0;
  }

  int bound_port = -1;
  while(true) {
    struct sockaddr_in subaddr;
    memset(&subaddr, 0, sizeof(subaddr));
    subaddr.sin_family = AF_INET;
    subaddr.sin_addr.s_addr = INADDR_ANY;
    int tport = rand() % (JOBCONN_LAST_PORT - JOBCONN_FIRST_PORT);
    tport += JOBCONN_FIRST_PORT;
    subaddr.sin_port = htons(tport);

    int brc = bind(subsock, (const struct sockaddr*)&subaddr, sizeof(struct sockaddr_in));
    if(brc == 0) {
      bound_port = tport;
      break;
    }
    log_info("port collission - retrying");
  }
  listen(subsock, 3);

  //
  // Set up pipe for receiving info from the subprocess handler
  //
  int subpipes[2];
  int prc = pipe(subpipes);
  if(prc == -1) {
    close(subsock);
    log_info("Cannot create pipes for subprocess handler comms!");
    return 0;
  }

  //
  // Now fork subprocess handler
  //
  int subpid = fork();
  if(subpid == -1) {
    close(subsock);
    log_info("Cannot fork subprocess handler!");
    return 0;
  }

  //
  // Child runs subprocess hadler
  //
  if(subpid == 0) {
    //
    // Re-open syslog connection
    //
    closelog();
    openlog("antsdph", LOG_NDELAY, LOG_DAEMON);

    //
    // Clean up stuff
    //
    close(subpipes[0]);

    //
    // Set default signal handler for SIGTERM
    //
    signal(SIGTERM, SIG_DFL);

    //
    // Fill in structure and run subprocess handler routine
    //
    runjob jobinfo;

    jobinfo.isock = subsock;
    jobinfo.peer = from;
    jobinfo.serial = serial;
    jobinfo.parent_fd = subpipes[1];
    jobinfo.username = username;
    jobinfo.type = type;

    subprocess_handler(jobinfo);
    exit(0);
  }

  //
  // We clean up after the subprocess setup
  //
  close(subsock);
  close(subpipes[1]);

  //
  // Insert subprocess handler in list of subprocesses
  //
  process pinfo;
  pinfo.type = type;
  pinfo.pid = subpid;
  pinfo.peer = from;
  pinfo.spawn_time = time(0);
  pinfo.auth_ok = false;

  process_table.insert(std::make_pair(subpipes[0], pinfo));

  return bound_port;
}
