//
// The sub-process handler
//
// This routine is running as a stand-alone process forwarding I/O
// between the running job and the remote rant process.
//
// $Id: processhandler.cc,v 1.7 2001/03/11 20:41:16 jakob Exp $
//
//  Copyright (C) 2001 Jakob Oestergaard <jakob@unthought.net>
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later
//  version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include "processes.hh"
#include "antsd.hh"

#include "protocol.h"

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#ifdef HAVE_SYSLOG_H
# include <syslog.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdio.h>
#include <pwd.h>
#include <cerrno>
#include <cstring>

#include <algorithm>


#ifdef USE_IDENTD_AUTH
 static char * ident_buffer = 0;
#endif

static char * buffer = 0;

static char ** args;

static int rant_sock = -1;
static std::string workdir;
static int fd_stdin = -1, fd_stdout = -1, fd_stderr = -1;
static int job_uid = -1, job_gid = -1;
static int job_pid = -1;

static void receive_rant(const address_t&, int&);
static void subprocess_auth(unsigned, const std::string&);
static void figure_out_uid(const std::string&);
static void spawn_that_sucker();
static void handle_messages();
static void emergency_bailout();
static void check_exit_status();

#ifdef DIRTY_SIGSEGV_DEBUG
void sigsegv_handler(int)
{
  log_info("CAUGHT SIGSEGV");
  raise(SIGSTOP);
}
#endif

void subprocess_handler(runjob& proc)
{
  int rc;
  buffer = new char[PROCESS_MESSAGE_LEN];

#ifdef DIRTY_SIGSEGV_DEBUG
  struct sigaction sigsegv_act;
  memset(&sigsegv_act, 0, sizeof(sigsegv_act));
  sigsegv_act.sa_handler = sigsegv_handler;
  sigaction(SIGSEGV, &sigsegv_act, 0);
#endif

  //
  // Receive remote ants connection
  //
  receive_rant(proc.peer, proc.isock);

  //
  // Authenticate request
  //
  subprocess_auth(proc.serial, proc.type);

  //
  // Receive name of executable and all the args...
  //
  rc = recv(rant_sock, buffer, PROCESS_MESSAGE_LEN-1, 0);
  if(rc <= 0) {
    log_info("Short read on process control connection when reading exename", 1);
    exit(1);
  }
  if(rc >= (signed)PROCESS_MESSAGE_LEN) {
    log_info("Can't happen !  Contact the author...", 1);
    exit(2);
  }
  buffer[rc] = 0;  // protects us from reading beyond end of buffer

  if(rc < 4) {
    log_info("Cannot use a string that's shorter than 4 characters...", 1);
    exit(1);
  }

  //
  // Figure out the UID
  //
  figure_out_uid(proc.username);

  //
  // This is where the fun ends - we now give up root privileges.
  //
  setregid(job_gid, job_gid);
  setreuid(job_uid, job_uid);

  //
  // We parse stuff from rant *after* we gave up root privs.
  //
  unsigned nargs;
  sscanf(buffer, "%4X", &nargs);
  char * bufp = buffer + 4;
  rc -= 4;

  args = new (char*)[nargs+1];

  for(unsigned i = 0; i != nargs; i++) {
    if(rc <= 0) {
      log_info("rc <= 0 and i != nargs - this shouldn't happen!", 1);
      exit(1);
    }
    // remember this
    args[i] = bufp;
    // move buffer head
    unsigned len = strlen(bufp) + 1;
    bufp += len;
    rc -= len;
  }

  args[nargs] = 0;

  //
  // Spawn that sucker
  //
  spawn_that_sucker();

  log_info(std::string("We just succesfully spawned a job (") + (args[0] ? : "N/A") + ") for user " + proc.username
  	   + "(" + proc.peer.toString() + ")", 1);

  //
  // Tell parent that we're green
  //
  unsigned msg = subprocess_ok_msg;
  write(proc.parent_fd, &msg, sizeof(msg));
  log_info("We notified the parent that auth is ok");

  //
  // Now, enter the message handling loop that forwards I/O between the job and rant
  //
  handle_messages();
}


static void receive_rant(const address_t& goodpeer, int& isock)
{
  //
  // We should listen for incoming connections on proc.isock
  // and only accept those from proc.peer
  //
  while(true) {
    struct sockaddr_in padr;
    socklen_t adrlen = sizeof(padr);
    int arc = accept(isock, reinterpret_cast<struct sockaddr*>(&padr), &adrlen);
    if(arc == -1) {
      log_info(std::string("accept() failed in subprocess handler - errno = \"")
	       + strerror(errno) + "\"", 1);
      exit(1);
    }
    if(adrlen != sizeof(struct sockaddr_in)) {
      log_info("Internal error - bad reply from accept()", 1);
      exit(1);
    }
    address_t peer(padr);
    //
    // See that the address matches our expected peer
    //
    if( goodpeer.inet_addr == peer.inet_addr ) {
      // Ok
      rant_sock = arc;
      break;
    }
    //
    // Ditch and log
    //
    log_info("We received a connection from " + peer.toString() + " instead of " 
	     + goodpeer.toString() + " - ditching connection!", 1);
    close(arc);
    listen(isock, 3);
  }

  //
  // Ok, we're done - now stop listening!
  //
  close(isock);
  isock = -1;

  log_info("Subprocess handler received good connection from " + goodpeer.toString());
}



static void subprocess_auth(unsigned exp_serial, const std::string& exp_type)
{
  //
  // Read in [type] [seq#] [working directory]
  //
  alloc_auth auth;
  int rc = recv(rant_sock, &auth, sizeof(auth), 0);
  if(rc != sizeof(auth)) {
    log_info("Short read from process control connection", 1);
    exit(1);
  }

  //
  // Jobtype verification
  //
  auth.type[TYPE_LEN-1] = 0;
  std::string type(auth.type);

  if(type != exp_type) {
    log_info("Got bad type (\"" + type + "\" not \"" + exp_type + "\") from peer", 1);
    exit(1);
  }

  //
  // See if serial # is valid
  //
  if(auth.serial != exp_serial) {
    log_info("Got bad serial# from peer", 1);
    exit(1);
  }

  //
  // Scan workdir
  //
  auth.workdir[CWD_LENGTH-1] = 0;
  workdir = std::string(auth.workdir);

  //
  // Send handshake
  //
  rc = send(rant_sock, "OK DUDE", 8, 0);
  if(rc == -1) {
    log_info("Remote rant processess barfed on handshake.", 1);
    exit(1);
  }
} 


static void figure_out_uid(const std::string& username)
{
  //
  // Find out what UID to use
  //
#ifdef USE_IDENTD_AUTH
  ident_buffer = new char[PROCESS_MESSAGE_LEN];

#error Missing stuff in sock.cc and jobnet.cc to enable identd authentication

  log_info("Starting identd negotiations");

  int identd = socket(AF_INET, SOCK_STREAM, 0);
  if(identd < 0) {
    log_info("Couldn't create socket for identd connection");
    exit(1);
  }
  struct sockaddr_in radr;
  radr.sin_family = AF_INET;
  radr.sin_addr.s_addr = proc.peer.inet_addr;
  radr.sin_port = IDENT_PORT;
  int irc = connect(identd, (struct sockaddr*)&radr, sizeof(radr));
  if(irc) {
    log_info("Couldn't connect to remote identd");
    exit(1);
  }
  irc = sprintf(ident_buffer, "%u, %u\r\n", ntohs(proc.peer.inet_port), ntohs(JOBNET_PORT));
  rc = send(identd, ident_buffer, irc, 0);
  if(irc != rc) {
    log_info("Couldn't send request to remote identd");
    close(identd);
    exit(1);
  }
  rc = recv(identd, ident_buffer, PROCESS_MESSAGE_LEN, 0);
  close(identd);
  if(rc <= 0) {
    log_info("Short read or read error from remote identd");
    exit(1);
  }
  int pos = 0;
  while(pos < rc && ident_buffer[pos] != ',') pos++;
  while(pos < rc && ident_buffer[pos] != ':') pos++; pos++;
  while(pos < rc && ident_buffer[pos] != ':') pos++; pos++;
  while(pos < rc && ident_buffer[pos] != ':') pos++; pos++;
  if(pos >= rc) {
    log_info("Cannot parse reply from identd (" +  std::string(ident_buffer, rc) + ")");          
    exit(2);
  }
  int epos = pos;
  while(epos < rc && ident_buffer[epos] 
	&& ident_buffer[epos] != '\r'
	&& ident_buffer[epos] != '\n') epos++;
  const std::string username(&ident_buffer[pos], epos-pos);

  log_info("Got username from identd: \"" + username + "\"");
#endif

  //
  // Check that username is reasonable
  //
  struct passwd * pwent = getpwnam(username.c_str());
  if(!pwent) {
    log_info("Couldn't find passwd entry for \"" + username + "\"", 1);
    exit(1);
  }
  if(pwent->pw_uid < PROCESS_MIN_UID) {
    log_info("Refusing to spawn process as \"" + username + "\"", 1);
    exit(1);
  }

  job_uid = pwent->pw_uid;
  job_gid = pwent->pw_gid;
}


static void spawn_that_sucker()
{
  //
  // Setup pipes
  //
  int pfd_stdin[2];
  int pfd_stdout[2];
  int pfd_stderr[2];
  int rc = !!pipe(pfd_stdin) || !!pipe(pfd_stdout) || !!pipe(pfd_stderr);
  if(rc) {
    log_info("Failed creating pipes for process", 1);
    exit(1);
  }

  //
  // Spawn shell with exename in it - setting correct UID
  //
  if(!(job_pid = fork())) {
    close(pfd_stdin[1]);
    close(pfd_stdout[0]);
    close(pfd_stderr[0]);
    dup2(pfd_stdin[0], 0);
    dup2(pfd_stdout[1], 1);
    dup2(pfd_stderr[1], 2);

    //
    // chdir() AFTER we set uid/gid !
    //
    rc = chdir(workdir.c_str());
    if(rc) {
      log_info("Couldn't change to proper working directory!", 1);
      exit(1);
    }

    //
    // Now run the command... (not in a shell)
    //
    execvp(args[0], args);
    log_info("execl failed: " + std::string(strerror(errno)), 1);
    exit(1);          
  }
  close(pfd_stdin[0]);
  close(pfd_stdout[1]);
  close(pfd_stderr[1]);
  fd_stdin = pfd_stdin[1];
  fd_stdout = pfd_stdout[0];
  fd_stderr = pfd_stderr[0];

  if(job_pid < 0) {
    log_info("Couldn't fork the job", 1);
    exit(1);
  }
}



static void handle_messages()
{
  //
  // We enter a select() loop, where we forward messages between the
  // spawned job and the remote rant.
  // Also, we handle signals sent from remote rans to the job,
  // and we set the nice-level.
  //
  while(true) {
    fd_set input;

    //
    // Set up fd sets for select()
    //
    FD_ZERO(&input);

    if(fd_stdout < 0 || fd_stderr < 0 || rant_sock < 0) {
      log_info("Bad FD encountered - this shouldn't happen!", 1);
      emergency_bailout();
    }
    
    FD_SET(fd_stdout, &input);
    int maxfd = fd_stdout;
    FD_SET(fd_stderr, &input);
    maxfd = std::max(maxfd, fd_stderr);

    FD_SET(rant_sock, &input);
    maxfd = std::max(maxfd, rant_sock);

    //
    // and select
    //
    int rc = select(maxfd + 1, &input, 0, 0, 0);

    //
    // If our job process died, we get errno==EINTR
    //
    if(rc == -1) {
      if(errno != EINTR) {
	log_info(std::string("select failed (errno=") + strerror(errno) + ")", 1);
	exit(2);
      }

      //
      // Handle job death
      //
      check_exit_status();
      continue;
    }

    //
    // See if we should read from rant
    //
    if(FD_ISSET(rant_sock, &input)) {

      struct {
	unsigned cmd;
	unsigned misc;
	char buffer[PROCESS_MESSAGE_LEN];
      } rantcommand;

      // read command from rant
      rc = recv(rant_sock, &rantcommand, PROCESS_MESSAGE_LEN+2*sizeof(unsigned), 0);
      if(rc <= 0) {
	log_info("Read error from rant - will kill job", 1);
	check_exit_status();
	emergency_bailout();
      }

      // send handshake
      send(rant_sock, "WE'RE STILL HERE", 16, 0);

      // Interpret
      switch(rantcommand.cmd) {
      case 0: // standard input
	// Sanity check
	if(rantcommand.misc > PROCESS_MESSAGE_LEN) {
	  log_info("Got bad message length from rant... Killing everything.", 1);
	  emergency_bailout();
	}
	rc = write(fd_stdin, rantcommand.buffer, rantcommand.misc);
	if(rc < 0) {
	  log_info("Failed writing to sub process... Killing everything.", 1);
	  emergency_bailout();
	}
	break;
      default:
	log_info("Got unknown command from rant - killing everything.", 1);
	emergency_bailout();
      }
    }

    //
    // See if we should read from the job process
    //
    if(FD_ISSET(fd_stdout, &input)) {

      struct {
	unsigned cmd;
	char buffer[PROCESS_MESSAGE_LEN];
      } reply;

      reply.cmd = 0;

      // Read message
      rc = read(fd_stdout, reply.buffer, PROCESS_MESSAGE_LEN - sizeof(unsigned));
      if(rc <= 0) {
	log_info("Read error / short read from process stdout");
	check_exit_status();
	if(rc < 0) {
	  log_info("Read error was an actual error: " + std::string(strerror(errno)), 1);
	  emergency_bailout();
	}
      }
      if(rc > 0) {
	//
	// Zero terminate buffer
	//
	if(rc >= static_cast<signed>(PROCESS_MESSAGE_LEN)) {
	  log_info("rc > PROCESS_MESSAGE_LEN - can't happen!", 1);
	  emergency_bailout();
	}
	reply.buffer[rc] = 0;

	// Forward message
	rc = send(rant_sock, &reply, sizeof(unsigned)+rc, 0);
	if(rc < 0) {
	  log_info("Cannot write to remote rant - will kill job", 1);
	  emergency_bailout();
	} 

	// Read handshake
	rc = recv(rant_sock, buffer, PROCESS_MESSAGE_LEN, 0);
	if(rc < 0) {
	  log_info("Read info from rant - will kill job", 1);
	  emergency_bailout();
	}
      }
    }

    //
    // See if we should read stderr from the job process
    //
    if(FD_ISSET(fd_stderr, &input)) {

      struct {
	unsigned cmd;
	char buffer[PROCESS_MESSAGE_LEN];
      } reply;

      reply.cmd = 1;

      // Read message
      rc = read(fd_stderr, reply.buffer, PROCESS_MESSAGE_LEN - sizeof(unsigned));
      if(rc <= 0) {
	log_info("Read error / short read from process stderr", 1);
	check_exit_status();
	if(rc < 0) 
	  emergency_bailout();
      }
      if(rc > 0) {
	//
	// Zero terminate buffer
	//
	if(rc >= static_cast<signed>(PROCESS_MESSAGE_LEN)) {
	  log_info("rc > PROCESS_MESSAGE_LEN in fd_stderr read - can't happen!", 1);
	  emergency_bailout();
	}
	reply.buffer[rc] = 0;

	// Forward message
	rc = send(rant_sock, &reply, rc+sizeof(unsigned), 0);
	if(rc < 0) {
	  log_info("Cannot write to remote rant - will kill job", 1);
	  emergency_bailout();
	} 

	// Read handshake
	rc = recv(rant_sock, buffer, PROCESS_MESSAGE_LEN, 0);
	if(rc < 0) {
	  log_info("Read info from rant - will kill job", 1);
	  emergency_bailout();
	}
      }
    }
  }
}


static void emergency_bailout()
{
  //
  // See if the process was already dead
  //
  int rc = waitpid(job_pid, 0, WNOHANG);
  if(rc != job_pid) {
    //
    // Send TERM signal
    //
    kill(job_pid, SIGTERM);
    //
    // Then wait for a little while
    //
    sleep(EMERGENCY_KILL_TIMEOUT);
    //
    // Now see if process is dead...
    //
    rc = waitpid(job_pid, 0, WNOHANG);
    //
    // If it's not - make damn sure it dies
    //
    if(rc != job_pid) {
      kill(job_pid, SIGKILL);
    }
  }
  log_info("Emergency bailout - job = " + std::string(args[0] ? args[0] : "N/A!"), 1);
  //
  // init will clean up after the job dies.
  // Our job is done.
  //
  exit(1);
}


static void check_exit_status()
{
  int status;
  int rc;

  //
  // If someone else died, we don't care
  //
  do {
    rc = waitpid(-1, &status, WNOHANG);
    if(rc == job_pid)
      break;
  } while(rc != 0);

  if(rc != job_pid)
    return;

  //
  // Our job process died
  //
  if(WIFEXITED(status)) {
    //
    // It exited normally - send back status
    //
    struct {
      unsigned cmd;
      int stat;
    } reply = { 2, WEXITSTATUS(status) };
	
    rc = send(rant_sock, &reply, sizeof(reply), 0);
    if(rc == -1) {
      log_info("Couldn't send back exit status to rant.", 1);
      exit(1);
    }	
  } else if(WIFSIGNALED(status)) {
    //
    // Process died due to a signal - send it back to rant
    //
    struct {
      unsigned cmd;
      int stat;
    } reply = { 3, WTERMSIG(status) };
	
    rc = send(rant_sock, &reply, sizeof(reply), 0);
    if(rc == -1) {
      log_info("Couldn't send back signal to rant.", 1);
      emergency_bailout();
    }	
  }
  //
  // Receive handshake
  //
  rc = recv(rant_sock, buffer, PROCESS_MESSAGE_LEN, 0);
  //
  // Exit
  //
  log_info("Exiting because job " + std::string(args[0] ? args[0] : "(N/A!)") + " ended.", 1);
  shutdown(rant_sock, 2);
  exit(0);
}
