/*
 * The Remote ANTs execution utility
 *
 * $Id: rant.c,v 1.9 2001/03/11 20:41:16 jakob Exp $
 *
 *  Copyright (C) 2000 Jakob Oestergaard <jakob@ostenfeld.dk>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later
 *  version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA
 *
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <netinet/in.h>
#include <netdb.h>
#include <netinet/tcp.h>
#include <time.h>
#include <arpa/inet.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "protocol.h"

#define ANTSD_SOCKET "/tmp/antsd"
#define MAX_ANTSD_STREAM  65535
#define MAX_ANTSD_MESSAGE 1024
#define RETRY_DELAY 500
#define JOBNET_PORT htons(86)
/*#define JOBNET_PORT htons(1086)*/
#define PRINT_INTERVAL 10

void usage(void);
void fail(const char*);

int setup_antsd_connection(void);

struct sockaddr_in jadr;
char type[TYPE_LEN];

int main(int argc, char** argv)
{
  char ** argp = argv + 1;
  int antsd;
  char buffer[MAX_ANTSD_STREAM];
  char * bufp;
  int buflen;
  int rc;

  if(argc == 1)
    usage();

  if(!strcmp("-t", *argp)) {
    argp++; argc -= 2;
    if(!*argp)
      fail("Cannot use -t and then nothing");
    if(strlen(*argp) >= TYPE_LEN)
      fail("Type too long");
    strcpy(type, *argp);
  } else {
    if(strlen(*argp) >= TYPE_LEN)
      fail("If you don't use a specific type, your command name\n"
           "will be used instead, and it should be short enough\n"
           "to fit as a type");
    strcpy(type, *argp);
    argp--; argc--;
  }

  signal(SIGPIPE, SIG_IGN);

  antsd = setup_antsd_connection();

  /* Read handshake before sending*/
  rc = recv(antsd, buffer, MAX_ANTSD_MESSAGE, 0);
  if(rc < 0)
    fail("Read error from remote antsd");

  /*
   * Now we should build a string holding the entire command
   * line we want to execute
   */
  bufp = buffer;
  buflen = 0;

  /*
   * That string is encoded like:
   *   0012[arg_0]\0[arg_1]\0....[arg_11]\0
   * four hex digits telling how many
   * arguments will follow, then the arguments
   * separated with \0 characters.
   */  

  bufp += sprintf(bufp, "%04X", argc - 1);

  while(*++argp) {
    buflen += strlen(*argp) + 1;
    if(buflen >= MAX_ANTSD_STREAM)
      fail("Argument list too long");
    memcpy(bufp, *argp, strlen(*argp) + 1);  /* + 1 is for the zero termination */
    bufp += strlen(*argp) + 1;  /* again, we want the zero termination in there */
  }

  rc = send(antsd, buffer, bufp - buffer, 0);
  if(rc != bufp - buffer) 
    fail("Short write to remote antsd");

  while((rc = recv(antsd, buffer, MAX_ANTSD_MESSAGE, 0)) > 0) {
    unsigned * cmd = (unsigned*)buffer;
    buffer[rc] = 0;
    switch(*cmd) {
    case 0:  /* stdout */
      fprintf(stdout, "%s", buffer + sizeof(unsigned));
      fflush(stdout);
      break;
    case 1:  /* stderr */
      fprintf(stderr, "%s", buffer + sizeof(unsigned));
      fflush(stderr);
      break;
    case 2: { /* exit - with status */
      char * p = buffer + sizeof(unsigned);
      int xit = *(int*)p;
      /* Send ACK */
      *cmd = 1; /* we're here still! */
      send(antsd, buffer, sizeof(unsigned), 0);
      /* Exit with status code */
      exit(xit);
    }
    break;
    case 3: { /* exit - with signal */
      char * p = buffer + sizeof(unsigned);
      int sig = *(int*)p;
      /* Send ACK */
      *cmd = 1; /* we're here still! */
      send(antsd, buffer, sizeof(unsigned), 0);
      /* Raise signal */
      fprintf(stderr, "rant: warning, remote process died from signal %i which I am about to raise too\n", sig);
      raise(sig);
      exit(-1);
    }
    break;
    default:
      fail("Got unknown reply from remote antsd");
    }
    *cmd = 1; /* we're here still! */
    rc = send(antsd, buffer, sizeof(unsigned), 0);
    if(rc != sizeof(unsigned))
      fail("Short send to remote antsd");
  }

  fprintf(stderr, "Remote antsd at %s closed connection (%s).\n",
	  inet_ntoa(jadr.sin_addr), strerror(errno));
  return 1;
}

void usage(void)
{
  fprintf(stderr, "Usage:\n"
          " spawn job:             rant [-n] [-t <type>] <command> <args...>\n"
          " query available slots: rant -q <type>\n"
          " Switches:  -n    don't transfer environment\n");
  exit(1);
}

void fail(const char* e) 
{
  fprintf(stderr, "Error:\n%s\n", e);
  exit(2);
}


int setup_antsd_connection(void)
{
  int antsd;
  int rc, src;
  struct alloc_reply response;
  struct alloc_auth auth;
  struct sockaddr_un addr;
  char cwd[CWD_LENGTH];
  int opt_val;
  struct linger lng;
  time_t last_print = time(0);


  do {
    struct antsd_request req;

    /* Connect and get a host suggestion */
    antsd = socket(AF_UNIX, SOCK_STREAM, 0);
    if(antsd == -1)
      fail("Cannot create AF_UNIX socket");

    addr.sun_family = AF_UNIX;
    strcpy(addr.sun_path, ANTSD_SOCKET);

    opt_val = 1;
    rc = setsockopt(antsd, SOL_SOCKET, SO_PASSCRED, &opt_val, sizeof(opt_val));
    if(rc)
      fail("Couldn't set SO_PASSCRED on antsd socket");

    rc = connect(antsd, (struct sockaddr*)&addr, sizeof(struct sockaddr_un));
    if(rc) 
      fail("Cannot connect to local antsd socket - antsd not running on localhost ?");

    req.request = 0;
    /* Tell antsd that it's a rant connection */
    rc = send(antsd, &req, sizeof(req), 0);
    if(rc != sizeof(req)) 
      fail("Cannot send request type to local antsd socket");

    /* Send request */
    rc = send(antsd, type, strlen(type)+1, 0);
    if(rc != strlen(type)+1) 
      fail("Cannot send to local antsd socket");

    rc = recv(antsd, &response, sizeof(response), 0);
    if(rc != sizeof(response))
      fail("Cannot read from local antsd socket");

    /* Responses:
     * 0  : No best host for the moment - ask again later
     * 1  : Connect to [addr] and submit job with seq# [seq]
     * 2  : Best host for the job is ourselves
     */

    if(response.status == 0) {
      time_t now = time(0);
      if(last_print + PRINT_INTERVAL < now) {
	fprintf(stderr, "(rant: Jobtype \"%s\" on hold, retrying...)\n", type);
	last_print = now;
      }
      close(antsd);
      usleep(1000 * RETRY_DELAY);
      continue;
    } else if(response.status == 1 || response.status == 2) {
      break;
    } else {
      fail("Got strange reply from local antsd");
    }
  } while(1);

  close(antsd);

  /*
   * Connect to antsd and authenticate
   */

  antsd = socket(AF_INET, SOCK_STREAM, 0);

  jadr.sin_family = AF_INET;
  jadr.sin_addr.s_addr = response.address;
  jadr.sin_port = htons(response.port);

  rc = connect(antsd, (struct sockaddr*)&jadr, sizeof(jadr));
  if(rc)
    fail("Cannot connect to remote antsd");

  opt_val = 1;
  rc = setsockopt(antsd, SOL_TCP, TCP_NODELAY, &opt_val, sizeof(opt_val));
  if(rc)
    fail("Cannot set TCP_NODELAY on remote antsd socket");

  lng.l_onoff = 0;
  lng.l_linger = 0;
  rc = setsockopt(antsd, SOL_SOCKET, SO_LINGER, &lng, sizeof(lng));
  if(rc)
    fail("Cannot disable SO_LINGER on remote antsd socket");

  /*
   * Send: [type] [seq] [working directory]
   */
  if(!getcwd(cwd, CWD_LENGTH))
    fail("Couldn't get current working directory");

  strncpy(auth.type, type, TYPE_LEN);
  auth.serial = response.serial;
  strncpy(auth.workdir, cwd, CWD_LENGTH);

  src = send(antsd, &auth, sizeof(auth), 0);
  if(src != sizeof(auth))
    fail("Error sending authentication to remote antsd");

  return antsd;
}
