//
// Local UNIX socket handling -- this is the command ``interpreter''
//
// $Id: sock.cc,v 1.15 2001/03/11 20:41:16 jakob Exp $
//
//  Copyright (C) 2000 Jakob Oestergaard <jakob@unthought.net>
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later
//  version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "sock.hh"
#include "antsd.hh"
#include "jobnet.hh"
#include "cluster.hh"
#include "processes.hh"

#include "protocol.h"

#include <sys/types.h>
#include <sys/socket.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <sys/un.h>
#include <stdio.h>
#include <sys/stat.h>
#include <pwd.h>
#include <errno.h>
#include <fcntl.h>

#include <map>
#include <list>
#include <algorithm>
#include <set>

int localsock = -1;

static void handle_jobslot_allocation(int, std::set<int>&);
static void handle_metrics_inquiry(int, std::set<int>&);

static void local_return_jobslot(unsigned, unsigned, unsigned, unsigned, const address_t&);

static std::set<int> connections;
static std::map<int,std::list<std::string> > pending_writes;

static std::map<address_t, std::map<unsigned, int> > pending_acknacks;

static std::set<int> metrics_update_clients;

//
// Whenever we want to write to a local connection - use this
// routine.
//
static void schedule_write(int fd, const std::string& str)
{
  if(connections.find(fd) == connections.end()) {
    log_info("Write schedule attempt to non-existing connection", 1);
    return;
  }
  if(pending_writes[fd].size() > UNIX_MAX_LINES_PENDING) {
    log_info("Ignoring old scheduled write - client isn't reading fast enough...", 1);
    pending_writes[fd].pop_front();
  }
  pending_writes[fd].push_back(str);
}

//
// For killing local connections - use this one.
//
static void remove_connection(int fd)
{
  close(fd);
  connections.erase(fd);
  pending_writes.erase(fd);
  metrics_update_clients.erase(fd);
}

//
// Setup of the local UNIX socket
//
void setup_sock()
{
  localsock = socket(AF_UNIX, SOCK_STREAM, 0);
  if(localsock == -1) {
    log_info("Sorry, cannot create local socket");
    return;
  }
  
  unlink(LOCALSOCK.c_str());
  struct sockaddr_un addr;
  addr.sun_family = AF_UNIX;
  strcpy(addr.sun_path, LOCALSOCK.c_str());
  int rc = bind(localsock, reinterpret_cast<struct sockaddr*>(&addr), sizeof(addr));
  if(rc) {
    log_info("Binding on local socket \"" + LOCALSOCK + "\" failed.");
    close(localsock);
    localsock = -1;
    return;
  }
  chmod(LOCALSOCK.c_str(), 0666);

  listen(localsock, 5);
}

//
// Read on the main UNIX socket - meaning, we're accepting
// a new client.
//
void read_sock()
{
  //
  // Accept a connection
  //
  int sock = accept(localsock, 0, 0);
  if(sock == -1) {
    log_info("Accepting connection from local socket failed");
    return;
  }

  listen(localsock, 5);

  connections.insert(sock);
  log_info("Accepted new connection on local socket");

  struct linger lng = {0, 0};
  int rc = setsockopt(sock, SOL_SOCKET, SO_LINGER, &lng, sizeof(lng));
  if(rc)
    log_info("Cannot unset SO_LINGER on socket");

  unsigned bint = 1;
  rc = setsockopt(sock, SOL_SOCKET, SO_PASSCRED, &bint, sizeof(bint));
  if(rc)
    log_info("Cannot set SO_PASSCRED on socket");
}


void except_sock()
{
  close(localsock);
  localsock = -1;
  log_info("Local socket shut down due to errors");
}


void maintain_local()
{
  
}


void insert_local_readers_set(fd_set& fds, int& maxfd)
{
  for(std::set<int>::const_iterator ci = connections.begin();
      ci != connections.end(); ci++) {
    FD_SET(*ci, &fds);
    maxfd = std::max(maxfd, *ci);
  }
}


void insert_local_writers_set(fd_set& fds, int& maxfd)
{
  for(std::map<int,std::list<std::string> >::const_iterator wi = pending_writes.begin();
      wi != pending_writes.end(); wi++) {
    //
    // If we have pending writes - insert connection
    //
    if(wi->second.size()) {
      FD_SET(wi->first, &fds);
      maxfd = std::max(maxfd, wi->first);
    }
  }
}


void read_local_set(const fd_set& fds)
{
  std::set<int> expired;
  for(std::set<int>::const_iterator ci = connections.begin();
      ci != connections.end(); ci++)
    if(FD_ISSET(*ci, &fds)) {
      //
      // See whether this is a rant connection, or a gants connection
      //
      // 0 :  rant  (job allocation)
      // 1 :  gants (metric/limits inquiry)
      //
      antsd_request request;
      int rc =  recv(*ci, &request, sizeof(request), 0);
      if(rc == -1 && errno == EAGAIN) {
	log_info("Odd -  we tried reading when socket wasn't ready...", 1);
	continue;
      }
      if(rc != sizeof(request)) {
	//
	// If it was an error - log it
	//
	if(rc == -1) {
	  log_info(std::string("Couldn't get request type from local socket: ") + strerror(errno), 1);
	} else if(rc != 0) {
	  log_info("Got odd reply length from rant - will ditch connection");
	}
	//
	// If not, then it was (probably) rc == 0
	//
	expired.insert(*ci);
	continue;
      }

      //
      // Handle request
      //
      switch(request.request) {
      case 0:  
	handle_jobslot_allocation(*ci, expired);
	break;
      case 1:
	handle_metrics_inquiry(*ci, expired);
	break;
      default:
	log_info("Invalid request received on local socket", 1);
      }

    }
  //
  // Kill off old connections
  //
  while(expired.begin() != expired.end()) {
    remove_connection(*expired.begin());
    expired.erase(expired.begin());
  }
}


void write_local_set(const fd_set& fds)
{
  std::set<int> expired;
  //
  // Step thru all local connections that have pending writes...
  //
  for(std::map<int,std::list<std::string> >::iterator wi
	= pending_writes.begin(); wi != pending_writes.end(); wi++)
    if(FD_ISSET(wi->first, &fds) && wi->second.size()) {
      //
      // Make sure socket is non-blocking
      //
      int flags = fcntl(wi->first, F_GETFL);
      if(flags == -1) {
	log_info("Cannot get flags on socket...", 1);
	continue;
      }
      fcntl(wi->first, F_SETFL, flags | O_NONBLOCK);

      //
      // Attempt write
      //
      std::string& cbuf = *wi->second.begin();
      int rc = send(wi->first, cbuf.data(), cbuf.size(), 0);
      if(rc == 0) {
	//
	// Hmm... Log it.
	//
	log_info("send returned 0 - most odd...", 1);

      } else if(rc > 0) {
	//
	// Remove what was sent
	//
	cbuf.erase(0, rc);
	//
	// Is anything left ?
	//
	if(!cbuf.size()) {
	  wi->second.pop_front();
	}

      } else if(rc < 0) {
	if(errno == EAGAIN) {
	  //
	  // Not much else to do than to wait.
	  //
	  log_info("send said try again later...", 1);
	} else {
	  //
	  // Log it - and kill the connection... Probably the client
	  // shut down the connection.
	  //
	  log_info(std::string("send failed : ") + strerror(errno), 1);
	  expired.insert(wi->first);
	}
      }
      
      //
      // Make socket blocking once again....
      //
      fcntl(wi->first, F_SETFL, flags);
    }
  //
  // Kill off old connections
  //
  while(expired.begin() != expired.end()) {
    remove_connection(*expired.begin());
    expired.erase(expired.begin());
  }
}


static void handle_jobslot_allocation(int ci, std::set<int>& expired)
{
  //
  // Get pid/gid/uid from the UN*X socket
  //
  unsigned pid, uid, gid;
  struct ucred creds;
  socklen_t clen = sizeof(struct ucred);
  int rc = getsockopt(ci, SOL_SOCKET, SO_PEERCRED, &creds, &clen);
  if(rc < 0) {
    log_info(std::string("Couldn't get credentials data from rant connection: ") + strerror(errno), 1);
    expired.insert(ci);
    return;
  }
  uid = creds.uid;
  gid = creds.gid;
  pid = creds.pid;

  //
  // Now look up username
  //
  struct passwd * pwent = getpwuid(uid);
  if(!pwent) {
    log_info("Invalid user id given in rant message", 1);
    expired.insert(ci);
    return;
  }
  const string username(pwent->pw_name);
  log_info("User \"" + username + "\" requested job slot");
      
  //
  // rant will send us a jobtype, 
  // we should then allocate a node and return the serial# and node address
  //
  char type[TYPE_LEN];
  rc = recv(ci, type, TYPE_LEN, 0);
  if(rc <= 0) {
    log_info("Read error on local connection. Removing.", 1);
    expired.insert(ci);
    return;
  }
  if(rc < TYPE_LEN) {
    type[rc] = 0;
  } else {
    type[TYPE_LEN-1] = 0;
  }
  std::string jtype = type;

  //
  // Find a good host for the job
  //
  address_t host;
  bool cando = suggest_job_allocation(jtype, host);
  if(!cando) {
    // Send back address 0
    log_info("There is no best host for job type \"" + jtype + "\" currently.");
    alloc_reply response;
    response.status = 0;
    response.serial = response.address = response.port = 0;
    send(ci, &response, sizeof(response), 0);
    return;
  }

  if(host == local_address) {

    log_info("Best host for job type \"" + jtype + "\" is ourselves right now");

    //
    // Allocate serial#
    //
    unsigned serial = register_allocation(jtype, username, local_address);
    if(!serial) {
      log_info("Error allocating job on local node", 1);
      alloc_reply response;
      response.status = 0;
      response.serial = response.address = response.port = 0;
      send(ci, &response, sizeof(response), 0);
      return;
    }

    //
    // Allocate port# and start up job handler process
    //
    unsigned portnum = startup_subprocess(local_address, serial, username, jtype);
    if(!portnum) {
      log_info("Error starting up jobhandler for local job", 1);
      alloc_reply response;
      response.status = 0;
      response.serial = response.address = response.port = 0;
      send(ci, &response, sizeof(response), 0);
      return;
    }

    //
    // Send back reply
    //
    alloc_reply response;
    response.status = 2;
    response.serial = serial;
    response.address = local_address.inet_addr;
    response.port = portnum;
    int rc = send(ci, &response, sizeof(response), 0);
    if(rc != sizeof(response)) {
      log_info("Write error when ack'ing local client request");
      return;
    }
  } else {
    log_info("Found a good host for job type \"" + jtype + "\"");
    //
    // Send job allocation request to remote node, register the sequence
    // number locally, register a callback for the sequence number
    //
    unsigned seqnr = request_remote_allocation(jtype, username, host);
    register_acknack_callback(seqnr, local_return_jobslot);
    pending_acknacks[host][seqnr] = ci;
  }
}


static void handle_metrics_inquiry(int ci, std::set<int>& expired)
{
  //
  // Ok - what we do is, we register this client as one that
  // wants to receive metric/limit updates.
  //
  
  metrics_update_clients.insert(ci);  

}

void notify_client_metrics_change(const address_t& peer, const std::string& jobtype, 
				  const jobt_limits& lim, const jobt_metrics& met)
{
  //
  // Send back the metrics and limits to all subscribers
  //
  // Format:
  //   <host> <job_type> <max> <pref> <running> <prealloc>
  //
  for(std::set<int>::const_iterator mi = metrics_update_clients.begin();
      mi != metrics_update_clients.end(); mi++) {

    char buf[LOG_LEN];
    memset(buf, 0, LOG_LEN);
    snprintf(buf, LOG_LEN, "%u %15s %u %u %u %u\n",
	     peer.inet_addr, jobtype.c_str(),
	     lim.max_instances, lim.preferred_instances,
	     met.instances, met.preallocations);

    schedule_write(*mi, std::string(buf, LOG_LEN));
  }
}


static void local_return_jobslot(unsigned ack, unsigned sequence, 
				 unsigned reason, unsigned extra, const address_t& from)
{
  //
  // We just received an ACK or a NACK from someone, let's see if it's
  // a response to a previous request_remote_allocation() request.
  //
  std::map<address_t, std::map<unsigned, int> >::iterator ai = pending_acknacks.find(from);
  if(ai == pending_acknacks.end()) {
    log_info("Got acknack from host we didn't expect one from");
    return;
  }
  
  std::map<unsigned, int>::iterator si = ai->second.find(sequence);
  if(si == ai->second.end()) {
    log_info("Got acknack sequence number we didn't expect");
    return;
  }

  if(connections.find(si->second) == connections.end()) {
    log_info("Oups, client no longer present, so we're dropping the acknack");
    ai->second.erase(si);
    return;
  }
  int client_fd = si->second;
  ai->second.erase(si);

  //
  // We either got an ack or a nack.  No matter what, we should remove
  // the registration retransmit handler running from jobnet.cc.
  //
  acknowledge_remote_allocation(sequence, from);

  //
  // If it was an ack, write back to client.
  //
  if(ack) {
    alloc_reply response;
    response.status = 1;
    response.serial = reason; // reason is used for serial#
    response.address = from.inet_addr;
    response.port = extra; // extra is used for port#
    int rc = send(client_fd, &response, sizeof(response), 0);
    if(rc != sizeof(response)) {
      log_info("Write error when ack'ing client request");
      return;
    }
    log_info("Ack'ed client request for job slot");
    return;
  }
  //
  // Somehow the best possible host denied us a job slot.
  // We should tell the client to try again.
  //
  // This could be optimized somewhat. We should request new metrics
  // from the best host, re-consider, retry allocation, etc.  This
  // approach works as well, but it may be slower.
  //
  alloc_reply response;
  response.status = 0;
  response.serial = response.address = response.port = 0;
  send(client_fd, &response, sizeof(response), 0);
  log_info("Told client that there is no best host at the moment");

  //
  // We request a metric update from the remote node to improve chances
  // that next client retry is considered with improved knowledge of the
  // current situation
  //
  request_new_metrics(from);
  request_new_limits(from);
}

void notify_client_alloc_failure(const address_t& host, unsigned sequence)
{
  std::map<address_t, std::map<unsigned, int> >::iterator ai = pending_acknacks.find(host);
  if(ai == pending_acknacks.end()) {
    log_info("Failure to allocate job slot on host we don't know");
    return;
  }
  std::map<unsigned, int>::iterator si = ai->second.find(sequence);
  if(si == ai->second.end()) {
    log_info("Failure to allocate job slot with sequence we don't know");
    return;
  }

  alloc_reply response;
  response.status = 0;
  response.serial = response.address = response.port = 0;
  send(si->second, &response, sizeof(response), 0);
  log_info("Told client to retry, as re-transmission of job slot allcation request failed");
  ai->second.erase(si);
  
}
